from elysia.objects import Text
from typing import List
from pydantic import BaseModel, Field


class TextWithCitation(BaseModel):
    text: str = Field(description="The text within the summary")
    ref_ids: List[str] = Field(
        description=(
            "The ref_ids of the citations relevant to the text. "
            "Can be an empty list if the text is not related to any of the citations."
        )
    )


class TextWithTitle(Text):
    def __init__(self, text: str, title: str, **kwargs):
        Text.__init__(
            self,
            "text_with_title",
            [{"text": text}],
            metadata={"title": title},
            **kwargs,
        )


class TextWithCitations(Text):
    def __init__(self, cited_texts: List[TextWithCitation], title: str, **kwargs):
        Text.__init__(
            self,
            "text_with_citations",
            [
                {
                    "text": cited_text_item.text,
                    "ref_ids": cited_text_item.ref_ids,
                }
                for cited_text_item in cited_texts
            ],
            metadata={"title": title},
            **kwargs,
        )



from elysia import Tool
from elysia.objects import Response

# Import a custom tool from a separate file
from elysia.tools.visualisation.linear_regression import BasicLinearRegression

# Import existing tools
from elysia.tools.retrieval.query import Query
from elysia.tools.retrieval.aggregate import Aggregate
from elysia.tools.text.text import CitedSummarizer, FakeTextResponse


# Or you can define the tool inline here
class TellAJoke(Tool):
    """
    Example tool for testing/demonstration purposes.
    Simply returns a joke as a text response that was an input to the tool.
    """

    def __init__(self, **kwargs):

        # Init requires initialisation of the super class (Tool)
        super().__init__(
            name="tell_a_joke",
            description="Displays a joke to the user.",
            inputs={
                "joke": {
                    "type": str,
                    "description": "A joke to tell.",
                    "required": True,
                }
            },
            end=True,
        )

    # Call must be a async generator function that yields objects to the decision tree
    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):

        # This example tool only returns the input to the tool, so is not very useful
        yield Response(inputs["joke"])

        # You can include more complex logic here via a custom function



import dspy


class ObjectSummaryPrompt(dspy.Signature):
    """
    Given a list of objects (a list of dictionaries, where each item in the dictionary is a field from the object),
    you must provide a list of strings, where each string is a summary of the object.

    These objects can be of any type, and you should summarise them in a way that is useful to the user.
    """

    objects: list[dict] = dspy.InputField(
        desc="The objects to summarise.", format=list[dict]
    )
    summaries: list[str] = dspy.OutputField(
        desc="""
        The summaries of each individual object, in a list of strings.
        These summaries should be extremely concise, and not aiming to summarise all objects, but a brief summary/description of _each_ object.
        Specifically, what the object is, what it is about to give the user a concise idea of the individual objects. A few sentences at most.
        Your output should be a list of strings in Python format, e.g. `["summary_1", "summary_2", ...]`.
        Do not enclose with ```python or ```. Just the list only.
        """.strip(),
        format=list[str],
    )



import dspy
from pydantic import BaseModel, Field


class CollectionSummariserPrompt(dspy.Signature):
    """
    You are an expert data analyst who provides summaries of datasets.
    Your task is to provide a summary of the data. This should be concise, one paragraph maximum and no more than 5 sentences.
    Do not calculate any statistics such as length, just describe the data.
    This is to inform the user about the data in the collection.
    Use markdown formatting (such as headers, bold, italics, lists, etc.) to make the summary more readable.
    """

    sample_size = dspy.InputField(
        description="""
    The number of objects in the data_sample, out of a total number of objects in the collection.
    """
    )
    data_sample = dspy.InputField(
        description="""
    A subset of the data to summarise. This will be a list of JSON objects. Each item is an individual item in the dataset and has the same fields.
    Because this is a subset, do not make definitive statements about _all_ of what the data contains.
    Instead, you can make statements like "this data includes..."
    """
    )
    data_fields = dspy.InputField(
        description="""
    The fields that exist in the data. This will be a list of field names.
    """
    )
    overall_summary: str = dspy.OutputField(
        description="""
        An overall summary of the dataset. Use markdown formatting (such as headers, bold, italics, lists, etc.) to make the summary more readable.
        A brief paragraph.
        """
    )
    relationships: str = dspy.OutputField(
        description="""
    A brief overview of the relationships between the fields. Mention any overlap or relevant points about how the fields are related. Around 2-4 sentences.
    """
    )
    structure: str = dspy.OutputField(
        description="""
    An overview of the structure of the dataset. How is it formatted. Around 1-3 sentences.
    """
    )
    irregularities: str = dspy.OutputField(
        description="""
    Any irregularities in the dataset that you think are important to know. Anything that could cause problems for the user when querying the data. Around 1-2 sentences.
    If there are no irregularities, you don't need to return anything. Say 'No irregularities found.' or something similar.
    """
    )
    field_descriptions: dict[str, str] = dspy.OutputField(
        description="""
    A dictionary of the field names and a description of the type of data in the field, be descriptive. Around 1-2 sentences per field.
    The keys of this dictionary MUST be the field names in the data_fields input.
    """
    )


class ReturnTypePrompt(dspy.Signature):
    """
    You are an expert at determining the type of data in a collection.
    Given some possible return types, you should choose the most appropriate ones for this data.
    Base this decision on the data fields and the collection summary.
    """

    collection_summary = dspy.InputField(
        desc="A description of the collection.", format=str
    )
    data_fields = dspy.InputField(
        desc="The fields that exist in the data and their data types.", format=dict
    )
    example_objects = dspy.InputField(
        desc="Example objects to help you understand the data.", format=list[dict]
    )
    possible_return_types = dspy.InputField(
        desc="The possible return types for the collection to choose from, a dictionary with key value pairs corresponding to the return type and its description.",
        format=dict,
    )
    return_types: list[str] = dspy.OutputField(
        desc="""
        All different types of return types that would suite this data.
        This should be a list of strings.
        You can choose as many as possible from the keys of possible_return_types.
        These should be ONLY the return types that are in possible_return_types, and nothing else.
        Your output should be formatted as a Python list and nothing else.
        If no return types are appropriate, you should return an empty list i.e. []. This possibility should be considered.
        """.strip(),
        format=str,
    )


class DataMappingPrompt(dspy.Signature):
    """
    You are an expert at mapping data fields to existing field names.
    """

    mapping_type: str = dspy.InputField(
        desc=(
            "The type of mapping to perform. "
            "Some examples are 'conversation', 'message', 'ticket', 'product', 'document', 'table', 'generic'. "
            "If the type is 'conversation', then the field mapping should include a 'conversation_id' field and a 'message_id' field. "
            "If there is no suitable mapping, return these fields as empty strings."
        ),
        format=str,
    )
    input_data_fields: list[str] = dspy.InputField(
        desc="The input fields to map.", format=list[str]
    )
    output_data_fields: list[str] = dspy.InputField(
        desc="The output fields to map to.", format=list[str]
    )
    input_data_types: dict[str, str] = dspy.InputField(
        desc="The data types of the input fields, a dictionary with key value pairs corresponding to the field and its data type.",
        format=dict[str, str],
    )
    collection_information: dict[str, str | dict] = dspy.InputField(
        desc="""
        Information about the collection.
        This is of the form:
        {
            "name": collection name,
            "length": number of objects in the collection,
            "summary": summary of the collection,
            "fields": [
                {
                    "name": field_name,
                    "type": the data type of the field.
                    "groups": a dict with the value and count of each group.
                        a comprehensive list of all unique values that exist in the field.
                        if this is None, then no relevant groups were found.
                        these values are string, but the actual values in the collection are the 'type' of the field.
                    "mean": mean of the field. if the field is text, this refers to the means length (in tokens) of the texts in this field. if the type is a list, this refers to the mean length of the lists,
                    "range": minimum and maximum values of the length.
                },
                ...
            ]
        }
        """.strip(),
        format=str,
    )
    example_objects: list[dict] = dspy.InputField(
        desc="Example objects to help you understand the data.", format=list[dict]
    )

    field_mapping: dict[str, str] = dspy.OutputField(
        desc="""
        Your output should be a dictionary of the form:
        {
            "input_field_name": "output_field_name",
            ...
        }
        E.g.
        {
            "title": "name",
            ...
        }
        would map the `title` field to the `name` field.
        If there is no mapping, you should include an empty string, e.g.
        {
            "title": "",
            ...
        }
        Complete all the fields, even if some of them are empty.
        But you can leave out certain output_field's if they are not needed.
        In summary, you should ALWAYS include all input fields as keys of the dictionary, but you can leave out certain output_field's if they are not needed.
        This should be formatted as a Python dictionary and nothing else.
        Do not create any new fields, only map existing fields.
        The keys in your output dictionary should be the input_field_names and no others.
        """.strip(),
        format=dict[str, str],
    )


class PromptSuggestorPrompt(dspy.Signature):
    """
    You are an expert at suggesting prompts for a given data collection.
    """

    collection_information: dict[str, str | dict] = dspy.InputField(
        desc="""
        Information about the collection.
        This is of the form:
        {
            "name": collection name,
            "length": number of objects in the collection,
            "summary": summary of the collection,
            "fields": [
                {
                    "name": field_name,
                    "groups": a dict with the value and count of each group.
                        a comprehensive list of all unique values that exist in the field.
                        if this is None, then no relevant groups were found.
                        these values are string, but the actual values in the collection are the 'type' of the field.
                    "mean": mean of the field. if the field is text, this refers to the means length (in tokens) of the texts in this field. if the type is a list, this refers to the mean length of the lists,
                    "range": minimum and maximum values of the length.
                    "type": the data type of the field.
                },
                ...
            ]
        }
        """.strip(),
        format=str,
    )
    example_objects: list[dict] = dspy.InputField(
        desc="Example objects to help you understand the data.", format=list[dict]
    )

    prompt_suggestions: list[str] = dspy.OutputField(
        desc="""
        A list of prompts that would be useful for the user to use to query the data.
        These should be in the form of a question that the user could ask to get information about the data.
        These prompts should be around 5-10 words long.
        They should show actual understanding of the data, and not just be a generic question.
        Use no formatting, just the plain text of the prompt.
        Aim to impress the user with your understanding of the data, and generate interesting and meaningful questions.
        Also aim to create questions that will help the user dive deeper into their data.
        Look for interactions between the fields, and connections that the user may not see themselves.
        Produce 10 questions.
        The prompts should either: specifically reference the collection, or be specific enough to the data that at a glance someone would recognise the data.
        """
    )



import inspect
import json
import time
import textwrap
from copy import deepcopy
from typing import AsyncGenerator, Literal

import dspy
from pympler import asizeof
from rich import print
from rich.console import Console
from rich.panel import Panel

import uuid

# Weaviate
import weaviate.classes.config as wc
from weaviate.util import generate_uuid5

# Elysia
from elysia.config import Settings, ElysiaKeyManager
from elysia.config import settings as environment_settings
from elysia.tree.util import (
    DecisionNode,
    TreeReturner,
    Decision,
    get_follow_up_suggestions,
    create_conversation_title,
)
from elysia.objects import (
    Completed,
    Result,
    Return,
    Update,
    Text,
    Tool,
    Warning,
    Error,
)
from elysia.tools.retrieval.aggregate import Aggregate
from elysia.tools.retrieval.query import Query
from elysia.tools.visualisation.visualise import Visualise
from elysia.tools.postprocessing.summarise_items import SummariseItems
from elysia.tools.text.text import (
    FakeTextResponse,
    CitedSummarizer,
)
from elysia.tree.util import ForcedTextResponse
from elysia.util.async_util import asyncio_run
from elysia.tree.objects import CollectionData, TreeData, Atlas, Environment
from elysia.util.client import ClientManager
from elysia.config import (
    check_base_lm_settings,
    check_complex_lm_settings,
    load_base_lm,
    load_complex_lm,
)
from elysia.util.objects import Tracker, TrainingUpdate, TreeUpdate
from elysia.util.parsing import remove_whitespace
from elysia.util.collection import retrieve_all_collection_names


class Tree:
    """
    The main class for the Elysia decision tree.
    Calling this method will execute the decision tree based on the user's prompt, and available collections and tools.
    """

    def __init__(
        self,
        branch_initialisation: Literal[
            "default", "one_branch", "multi_branch", "empty"
        ] = "default",
        style: str = "No style provided.",
        agent_description: str = "No description provided.",
        end_goal: str = "No end goal provided.",
        user_id: str | None = None,
        conversation_id: str | None = None,
        low_memory: bool = False,
        use_elysia_collections: bool = True,
        settings: Settings | None = None,
    ) -> None:
        """
        Args:
            branch_initialisation (str): The initialisation method for the branches,
                currently supports some pre-defined initialisations: "multi_branch", "one_branch".
                Set to "empty" to start with no branches and to add them, and the tools, yourself.
            style (str): The writing style of the agent. Automatically set for "multi_branch" and "one_branch" initialisation, but overrided if non-empty.
            agent_description (str): The description of the agent. Automatically set for "multi_branch" and "one_branch" initialisation, but overrided if non-empty.
            end_goal (str): The end goal of the agent. Automatically set for "multi_branch" and "one_branch" initialisation, but overrided if non-empty.
            user_id (str): The id of the user, e.g. "123-456",
                unneeded outside of user management/hosting Elysia app
            conversation_id (str): The id of the conversation, e.g. "123-456",
                unneeded outside of conversation management/hosting Elysia app
            low_memory (bool): Whether to run the tree in low memory mode.
                If True, the tree will not load the (dspy) models within the tree.
                Set to False for normal operation.
            use_elysia_collections (bool): Whether to use weaviate collections as processed by Elysia.
                If False, the tree will not use the processed collections.
            settings (Settings): The settings for the tree, an object of elysia.Settings.
                This is automatically set to the environment settings if not provided.
        """
        # Define base variables of the tree
        if user_id is None:
            self.user_id = str(uuid.uuid4())
        else:
            self.user_id = user_id

        if conversation_id is None:
            self.conversation_id = str(uuid.uuid4())
        else:
            self.conversation_id = conversation_id

        if settings is None:
            self.settings = environment_settings
        else:
            assert isinstance(
                settings, Settings
            ), "settings must be an instance of Settings"
            self.settings = settings

        self.use_elysia_collections = use_elysia_collections

        # Initialise some tree variables
        self.decision_nodes: dict[str, DecisionNode] = {}
        self.decision_history = [[]]
        self.tree_index = -1
        self.suggestions = []
        self.actions_called = {}
        self.query_id_to_prompt = {}
        self.prompt_to_query_id = {}
        self.retrieved_objects = []
        self.store_retrieved_objects = False
        self.conversation_title = None
        self.low_memory = low_memory
        self._base_lm = None
        self._complex_lm = None
        self._config_modified = False
        self.root = None

        # Define the inputs to prompts
        self.tree_data = TreeData(
            environment=Environment(),
            collection_data=CollectionData(
                collection_names=[], logger=self.settings.logger
            ),
            atlas=Atlas(
                style=style,
                agent_description=agent_description,
                end_goal=end_goal,
            ),
            recursion_limit=5,
            settings=self.settings,
        )

        # initialise the timers
        self.tracker = Tracker(
            tracker_names=["decision_node"],
            logger=self.settings.logger,
        )

        # Set the initialisations
        self.tools = {}
        self.set_branch_initialisation(branch_initialisation)
        self.tree_data.atlas.style = style
        self.tree_data.atlas.agent_description = agent_description
        self.tree_data.atlas.end_goal = end_goal

        self.tools["forced_text_response"] = ForcedTextResponse()

        # some variables for storing feedback
        self.action_information = []
        self.history = {}
        self.training_updates = []

        # -- Get the root node and construct the tree
        self._get_root()
        self.tree = {}
        self._construct_tree(self.root, self.tree)

        # initialise the returner (for frontend)
        self.returner = TreeReturner(
            user_id=self.user_id,
            conversation_id=self.conversation_id,
        )

        # Print the tree if required
        self.settings.logger.debug(
            "Initialised tree with the following decision nodes:"
        )
        for decision_node in self.decision_nodes.values():
            self.settings.logger.debug(
                f"  - [magenta]{decision_node.id}[/magenta]: {list(decision_node.options.keys())}"
            )

    @property
    def base_lm(self) -> dspy.LM:
        if self.low_memory:
            return load_base_lm(self.settings)
        else:
            if self._base_lm is None:
                self._base_lm = load_base_lm(self.settings)
            return self._base_lm

    @property
    def complex_lm(self) -> dspy.LM:
        if self.low_memory:
            return load_complex_lm(self.settings)
        else:
            if self._complex_lm is None:
                self._complex_lm = load_complex_lm(self.settings)
            return self._complex_lm

    def multi_branch_init(self) -> None:
        self.add_branch(
            root=True,
            branch_id="base",
            instruction="""
            Choose a base-level task based on the user's prompt and available information.
            You can search, which includes aggregating or querying information - this should be used if the user needs (more) information.
            You can end the conversation by choosing text response, or summarise some retrieved information.
            Base your decision on what information is available and what the user is asking for - you can search multiple times if needed,
            but you should not search if you have already found all the information you need.
            """,
            status="Choosing a base-level task...",
        )
        self.add_tool(branch_id="base", tool=CitedSummarizer)
        self.add_tool(branch_id="base", tool=FakeTextResponse)

        self.add_branch(
            root=False,
            branch_id="search",
            from_branch_id="base",
            instruction="""
            Choose between querying the knowledge base via semantic/keyword search, or aggregating information by performing operations, on the knowledge base.
            Querying is when the user is looking for specific information related to the content of the dataset, requiring a specific search query. This is for retrieving specific information via a _query_, similar to a search engine.
            Aggregating is when the user is looking for a specific operations on the dataset, such as summary statistics of the quantity of some items. Aggregation can also include grouping information by some property and returning statistics about the groups.
            """,
            description=f"""
            Search the knowledge base. This should be used when the user is lacking information for this particular prompt. This retrieves information only and provides no output to the user except the information.
            Choose to query (semantic or keyword search on a knowledge base), or aggregate information (calculate properties/summary statistics/averages and operations on the knowledge bases).
            """,
            status="Searching the knowledge base...",
        )
        self.add_tool(branch_id="search", tool=Query, summariser_in_tree=True)
        self.add_tool(branch_id="search", tool=Aggregate)
        self.add_tool(branch_id="base", tool=Visualise)
        self.add_tool(SummariseItems, branch_id="search", from_tool_ids=["query"])

    def one_branch_init(self) -> None:
        self.add_branch(
            root=True,
            branch_id="base",
            instruction="""
            Choose a base-level task based on the user's prompt and available information.
            Decide based on the tools you have available as well as their descriptions.
            Read them thoroughly and match the actions to the user prompt.
            """,
            status="Choosing a base-level task...",
        )
        self.add_tool(branch_id="base", tool=CitedSummarizer)
        self.add_tool(branch_id="base", tool=FakeTextResponse)
        self.add_tool(branch_id="base", tool=Aggregate)
        self.add_tool(branch_id="base", tool=Query, summariser_in_tree=True)
        self.add_tool(branch_id="base", tool=Visualise)
        self.add_tool(SummariseItems, branch_id="base", from_tool_ids=["query"])

    def empty_init(self) -> None:
        self.add_branch(
            root=True,
            branch_id="base",
            instruction="""
            Choose a base-level task based on the user's prompt and available information.
            Decide based on the tools you have available as well as their descriptions.
            Read them thoroughly and match the actions to the user prompt.
            """,
            status="Choosing a base-level task...",
        )

    def clear_tree(self) -> None:
        self.decision_nodes = {}
        self.root = None

    def set_branch_initialisation(self, initialisation: str | None) -> None:
        self.clear_tree()

        if (
            initialisation is None
            or initialisation == ""
            or initialisation == "one_branch"
            or initialisation == "default"
        ):
            self.one_branch_init()
        elif initialisation == "multi_branch":
            self.multi_branch_init()
        elif initialisation == "empty":
            self.empty_init()
        else:
            raise ValueError(f"Invalid branch initialisation: {initialisation}")

        self.branch_initialisation = initialisation

    def smart_setup(self) -> None:
        """
        Configures the `settings` object of the tree with the `Settings.smart_setup()` method.
        """

        self.settings = deepcopy(self.settings)
        self.settings.SETTINGS_ID = str(uuid.uuid4())
        self._config_modified = True
        self.settings.smart_setup()

    def configure(self, **kwargs) -> None:
        """
        Configure the tree with new settings.
        Wrapper for the settings.configure() method.
        Will not affect any settings preceding this (e.g. in TreeManager).
        """
        self.settings = deepcopy(self.settings)
        self.settings.SETTINGS_ID = str(uuid.uuid4())
        self._config_modified = True
        self.tree_data.settings = self.settings
        self.settings.configure(**kwargs)

    def change_style(self, style: str) -> None:
        self.tree_data.atlas.style = style
        self._config_modified = True

    def change_agent_description(self, agent_description: str) -> None:
        self.tree_data.atlas.agent_description = agent_description
        self._config_modified = True

    def change_end_goal(self, end_goal: str) -> None:
        self.tree_data.atlas.end_goal = end_goal
        self._config_modified = True

    def _get_root(self) -> None:
        for decision_node in self.decision_nodes.values():
            if decision_node.root:
                if self.root is not None and self.root != decision_node.id:
                    raise ValueError("Multiple root decision nodes found")

                self.root = decision_node.id

        if self.root is None:
            raise ValueError("No root decision node found")

    def _construct_tree(
        self, node_id: str | None, tree: dict, branch: bool = True
    ) -> dict:
        if node_id is None:
            raise ValueError("Node ID is None")

        decision_node = self.decision_nodes[node_id]

        # Ensure the order of the keys in each option is the same
        key_order = [
            "name",
            "id",
            "description",
            "instruction",
            "reasoning",
            "branch",
            "options",
        ]

        # Set the base node information
        tree["name"] = node_id.capitalize().replace("_", " ")
        tree["id"] = node_id
        if node_id == self.root:
            tree["description"] = ""
        tree["instruction"] = remove_whitespace(
            decision_node.instruction.replace("\n", "")
        )
        tree["reasoning"] = ""
        tree["branch"] = branch
        tree["options"] = {}

        # Order the top-level dictionary
        tree = {key: tree[key] for key in key_order if key in tree}

        # Initialize all options first with ordered dictionaries
        for option in decision_node.options:
            tree["options"][option] = {
                "description": remove_whitespace(
                    str(decision_node.options[option]["description"]).replace("\n", "")
                )
            }

        # Then handle the recursive cases
        for option in decision_node.options:
            next_node: DecisionNode | None = decision_node.options[option]["next"]  # type: ignore
            if (
                decision_node.options[option]["action"] is not None
                and next_node is None
            ):
                tree["options"][option]["name"] = option.capitalize().replace("_", " ")
                tree["options"][option]["id"] = option
                tree["options"][option]["instruction"] = ""
                tree["options"][option]["reasoning"] = ""
                tree["options"][option]["branch"] = False
                tree["options"][option]["options"] = {}

            elif next_node is not None:
                tree["options"][option] = self._construct_tree(
                    next_node.id,
                    tree["options"][option],
                    branch=decision_node.options[option]["action"] is None,
                )
            else:
                tree["options"][option]["name"] = option.capitalize().replace("_", " ")
                tree["options"][option]["id"] = option
                tree["options"][option]["instruction"] = ""
                tree["options"][option]["reasoning"] = ""
                tree["options"][option]["branch"] = True
                tree["options"][option]["options"] = {}

            # Order each option's dictionary
            tree["options"][option] = {
                key: tree["options"][option][key]
                for key in key_order
                if key in tree["options"][option]
            }

        return tree

    async def set_collection_names(
        self,
        collection_names: list[str],
        client_manager: ClientManager,
    ) -> None:
        self.settings.logger.debug(
            f"Using the following collection names: {collection_names}"
        )

        collection_names = await self.tree_data.set_collection_names(
            collection_names, client_manager
        )

    def _remove_empty_branches(self) -> list[str]:
        empty_branches = []
        for branch_id, branch in self.decision_nodes.items():
            if len(branch.options) == 0:
                empty_branches.append(branch_id)

        for branch_id in self.decision_nodes:
            for empty_branch in empty_branches:
                self.decision_nodes[branch_id].remove_option(empty_branch)

        for empty_branch in empty_branches:
            if empty_branch != self.root:
                self.settings.logger.warning(
                    f"Removing empty branch: {empty_branch} "
                    "No tools are attached to this branch, so it has been removed. "
                    f"To add a tool to this branch, use .add_tool(tool_name, branch_id={empty_branch})"
                )
                del self.decision_nodes[empty_branch]

        return empty_branches

    def _get_function_inputs(self, tool_name: str, inputs: dict) -> dict:
        if tool_name in self.tools:
            # any non-provided inputs are set to the default
            default_inputs = self.tools[tool_name].get_default_inputs()
            for default_input_name in default_inputs:
                if default_input_name not in inputs:
                    inputs[default_input_name] = default_inputs[default_input_name]

            # if the inputs match the 'schema' of keys: description, type, default, value, then take the value
            for input_name in inputs:
                if (
                    isinstance(inputs[input_name], dict)
                    and "value" in inputs[input_name]
                ):
                    inputs[input_name] = inputs[input_name]["value"]

            return inputs
        else:
            return {}

    async def _check_rules(
        self, branch_id: str, client_manager: ClientManager
    ) -> tuple[list[str], dict]:
        branch = self.decision_nodes[branch_id]
        nodes_with_rules_met = []
        rule_tool_inputs = {}
        for function_name, option in branch.options.items():
            if function_name not in self.tools:
                pass
            elif "run_if_true" in dir(self.tools[function_name]):
                rule_met, rule_tool_inputs = await self.tools[
                    function_name
                ].run_if_true(
                    tree_data=self.tree_data,
                    client_manager=client_manager,
                    base_lm=self.base_lm,
                    complex_lm=self.complex_lm,
                )
                if rule_met:
                    nodes_with_rules_met.append(function_name)
                    if rule_tool_inputs is None or rule_tool_inputs == {}:
                        rule_tool_inputs[function_name] = self.tools[
                            function_name
                        ].get_default_inputs()
                    else:
                        rule_tool_inputs[function_name] = rule_tool_inputs

        return nodes_with_rules_met, rule_tool_inputs

    def set_conversation_id(self, conversation_id: str) -> None:
        self.conversation_id = conversation_id
        self.returner.conversation_id = conversation_id

    def set_user_id(self, user_id: str) -> None:
        self.user_id = user_id
        self.returner.user_id = user_id

    def soft_reset(self) -> None:
        # conversation history is not reset
        # environment is not reset
        if self.low_memory:
            self.history = {}

        self.recursion_counter = 0
        self.tree_data.num_trees_completed = 0
        self.decision_history = [[]]
        self.training_updates = []
        self.tree_data.soft_reset()
        self.action_information = []
        self.tree_index += 1
        self.retrieved_objects = []
        self.returner.set_tree_index(self.tree_index)

    def save_history(self, query_id: str, time_taken_seconds: float) -> None:
        """
        What the tree did, results for saving feedback.
        """
        training_update = deepcopy(
            [update.to_json() for update in self.training_updates]
        )

        self.history[query_id] = {
            "num_trees_completed": self.tree_data.num_trees_completed,
            "tree_data": deepcopy(self.tree_data),
            "action_information": deepcopy(self.action_information),
            "decision_history": [
                item for sublist in deepcopy(self.decision_history) for item in sublist
            ],
            "base_lm_used": self.settings.BASE_MODEL,
            "complex_lm_used": self.settings.COMPLEX_MODEL,
            "time_taken_seconds": time_taken_seconds,
            "training_updates": training_update,
            "initialisation": f"{self.branch_initialisation}",
        }
        # can reset training updates now
        self.training_updates = []

    def set_start_time(self) -> None:
        self.start_time = time.time()

    def add_tool(
        self,
        tool,
        branch_id: str | None = None,
        from_tool_ids: list[str] = [],
        root: bool = False,
        **kwargs,
    ) -> None:
        """
        Add a Tool to a branch or on top of an existing tool.
        The tool needs to be an instance of the Tool class.

        Args:
            tool (Tool): The tool to add
            branch_id (str): The id of the branch to add the tool to
                If not specified, the tool will be added to the root branch
            from_tool_ids (list[str]): The ids of the tools to add the new tool after
                If not specified, the tool will be added to the base of the branch
            root (bool): Whether the tool is the root tool
                If not specified, the tool will be added to the root branch
            kwargs (any): Additional keyword arguments to pass to the initialisation of the tool

        Example 1:
            To add a tool, `Query`, to a branch called 'search', you can do this:
            ```python
            tree.add_tool(Query, branch_id="search")
            ```
            This will add the `Query` tool to the branch 'search'.
            If the branch 'search' doesn't exist, it will raise an error.
            To add a branch, use the `.add_branch()` method.


        Example 2:
            Assume your tree has a "search" branch with two tools: 'query' and 'aggregate'.
            You can add a tool, `CheckResult`, after the 'query' tool like this:
            ```python
            tree.add_tool(CheckResult, branch_id="search", from_tool_ids=["query"])
            ```
            This will add the `CheckResult` tool to the "search" branch, after the 'query' tool.
            So the "search" branch will still only have two options: 'query' and 'aggregate'.
            But after 'query', there will be a new option for the `CheckResult` tool.

        Example 3:
            You can add a tool, `SendEmail`, after the `CheckResult` (from Example 2) tool like this:
            ```python
            tree.add_tool(SendEmail, from_tool_ids=["query", "check_result"], root=True)
            ```
            It will add an additional option to the root branch, after the 'query' and 'check_result' tools.
        """

        if (
            inspect.getfullargspec(tool.__init__).varkw is None
            or inspect.getfullargspec(tool.__call__).varkw is None
        ):
            raise TypeError("tool __init__ and __call__ must accept **kwargs")

        if not inspect.isasyncgenfunction(tool.__call__):
            raise TypeError(
                "__call__ must be an async generator function. "
                "I.e. it must yield objects."
            )

        if isinstance(tool, Tool):
            tool_instance = tool
        else:
            tool_instance = tool(
                logger=self.settings.logger,
                **kwargs,
            )

        if not isinstance(tool_instance, Tool):
            raise TypeError("tool must be an instance of the Tool class")

        if "__call__" not in dir(tool_instance):
            raise TypeError("tool must be callable (have a __call__ method)")

        if "__init__" not in dir(tool_instance):
            raise TypeError("tool must have an __init__ method")

        if hasattr(tool_instance, "is_tool_available"):
            if not inspect.iscoroutinefunction(tool_instance.is_tool_available):
                raise TypeError(
                    "is_tool_available must be an async function that returns a single boolean value"
                )

        if hasattr(tool_instance, "run_if_true"):
            if not inspect.iscoroutinefunction(tool_instance.run_if_true):
                raise TypeError(
                    "run_if_true must be an async function that returns a single boolean value"
                )

        if root:
            if branch_id is not None:
                self.settings.logger.warning(
                    f"In .add_tool(), `root` is True, so `branch_id` ('{branch_id}') will be ignored. "
                    f"Tool: '{tool_instance.name}' will be added to the root branch ('{self.root}')."
                )
            branch_id = self.root

        if branch_id is None:
            branch_id = self.root

        if branch_id not in self.decision_nodes:
            raise ValueError(
                f"Branch '{branch_id}' not found. Use .add_branch() to add a branch before adding a tool. "
                f"Or, set `root=True` to add the tool to the root branch ('{self.root}')."
            )

        current_decision_node = self.decision_nodes[branch_id]
        for from_tool_id in from_tool_ids:
            if isinstance(current_decision_node, DecisionNode):
                if from_tool_id not in current_decision_node.options:
                    raise ValueError(
                        f"Tool '{from_tool_id}' not found in branch '{branch_id}'. "
                        f"Available options are: {list(current_decision_node.options.keys())}"
                    )

                current_decision_node = current_decision_node.options[from_tool_id][
                    "next"
                ]

        self.tools[tool_instance.name] = tool_instance

        if from_tool_ids == []:
            self.decision_nodes[branch_id].add_option(
                id=tool_instance.name,
                description=tool_instance.description,
                inputs=tool_instance.inputs,
                action=self.tools[tool_instance.name],
                end=tool_instance.end,
                status=tool_instance.status,
            )
        else:

            new_branch_id = branch_id
            for from_tool_id in from_tool_ids:
                new_branch_id += f".{from_tool_id}"

            # only create a new decision node if one doesn't exist here
            if new_branch_id not in self.decision_nodes:
                decision_node = DecisionNode(
                    id=new_branch_id,
                    instruction=f"Choose one of the actions based on their descriptions and the user prompt.",
                    options={},
                    root=False,
                    logger=self.settings.logger,
                    use_elysia_collections=self.use_elysia_collections,
                )
                self.decision_nodes[new_branch_id] = decision_node

                prev_branch_id = branch_id
                for from_tool_id in from_tool_ids[:-1]:
                    prev_branch_id += f".{from_tool_id}"

                self.decision_nodes[prev_branch_id].options[from_tool_ids[-1]][
                    "next"
                ] = self.decision_nodes[new_branch_id]

            # add the tool to the new decision node
            self.decision_nodes[new_branch_id].add_option(
                id=tool_instance.name,
                description=tool_instance.description,
                inputs=tool_instance.inputs,
                action=self.tools[tool_instance.name],
                end=tool_instance.end,
                status=tool_instance.status,
            )

        self.tracker.add_tracker(tool_instance.name)

        # reconstruct tree
        self._get_root()
        self.tree = {}
        self._construct_tree(self.root, self.tree)

    def remove_tool(
        self,
        tool_name: str,
        branch_id: str | None = None,
        from_tool_ids: list[str] = [],
        root: bool = False,
    ) -> None:
        """
        Remove a Tool from a branch.

        Args:
            tool_name (str): The name of the tool to remove.
            branch_id (str): The id of the branch to remove the tool from,
                if not specified, the tool will be removed from the root branch.
            from_tool_ids (list[str]): The ids of the tools to which precedes the tool to remove.
            root (bool): Whether the branch the tool is in is the root branch.
        """
        if root:
            if branch_id is not None:
                self.settings.logger.warning(
                    f"In .add_tool(), `root` is True, so `branch_id` ('{branch_id}') will be ignored. "
                    f"Tool: '{tool_name}' will be removed from the root branch ('{self.root}')."
                )
            branch_id = self.root

        if branch_id is None:
            branch_id = self.root

        if branch_id not in self.decision_nodes:
            raise ValueError(f"Branch {branch_id} not found.")

        if (
            tool_name not in self.decision_nodes[branch_id].options
            and from_tool_ids == []
        ):
            raise ValueError(f"Tool {tool_name} not found in branch {branch_id}.")

        current_decision_node = self.decision_nodes[branch_id]
        for from_tool_id in from_tool_ids:
            if isinstance(current_decision_node, DecisionNode):
                if from_tool_id not in current_decision_node.options:
                    raise ValueError(
                        f"Tool '{from_tool_id}' not found in branch '{current_decision_node.id}'. "
                        f"Available options are: {list(current_decision_node.options.keys())}"
                    )
                current_decision_node = current_decision_node.options[from_tool_id][
                    "next"
                ]

        if (
            isinstance(current_decision_node, DecisionNode)
            and tool_name not in current_decision_node.options
        ):
            raise ValueError(
                f"Tool '{tool_name}' not found in branch '{current_decision_node.id}'. "
                f"Available options are: {list(current_decision_node.options.keys())}"
            )

        if from_tool_ids == []:
            self.decision_nodes[branch_id].remove_option(tool_name)
        else:
            tool_branch_id = branch_id
            for from_tool_id in from_tool_ids:
                tool_branch_id += f".{from_tool_id}"
            tool_branch_id += f".{tool_name}"

            prev_branch_id = branch_id
            for from_tool_id in from_tool_ids:
                prev_branch_id += f".{from_tool_id}"

            self.decision_nodes[prev_branch_id].remove_option(tool_name)
            if self.decision_nodes[prev_branch_id].options == {}:
                del self.decision_nodes[prev_branch_id]
                stem_branch_id = prev_branch_id[: prev_branch_id.rfind(".")]
                for stem_branch_option in self.decision_nodes[
                    stem_branch_id
                ].options.values():
                    if (
                        stem_branch_option["next"] is not None
                        and isinstance(stem_branch_option["next"], DecisionNode)
                        and stem_branch_option["next"].id == prev_branch_id
                    ):
                        stem_branch_option["next"] = None

            if (
                tool_branch_id in self.decision_nodes
                and self.decision_nodes[tool_branch_id].options != {}
            ):
                self.settings.logger.warning(
                    f"The following tools stem from '{tool_branch_id}', "
                    f"and have also been removed: {list(self.decision_nodes[tool_branch_id].options.keys())}"
                )

            # find any decision nodes that stem from this
            nodes_to_remove = []
            for decision_node_id in self.decision_nodes:
                if decision_node_id.startswith(tool_branch_id):
                    if decision_node_id != tool_branch_id:
                        self.settings.logger.warning(
                            f"Decision node '{decision_node_id}' stems from '{tool_branch_id}'. "
                            f"Removing tool '{tool_name}' has also removed '{decision_node_id}'."
                        )
                    nodes_to_remove.append(decision_node_id)

            for decision_node_id in nodes_to_remove:
                del self.decision_nodes[decision_node_id]

        del self.tools[tool_name]
        self.tracker.remove_tracker(tool_name)

        # reconstruct tree
        self._get_root()
        self.tree = {}
        self._construct_tree(self.root, self.tree)

    def add_branch(
        self,
        branch_id: str,
        instruction: str,
        description: str = "",
        root: bool = False,
        from_branch_id: str = "",
        from_tool_ids: list[str] = [],
        status: str = "",
    ) -> None:
        """
        Add a branch to the tree.

        args:
            branch_id (str): The id of the branch being added.
            instruction (str): The general instruction for the branch, what is this branch containing?
                What kind of tools or actions are being decided on this branch?
                Only displayed to the decision maker when this branch is chosen.
            description (str): A description of the branch, if it is to be chosen from a previous branch.
                How does the model know whether to choose this branch or not?
            root (bool): Whether this is the root branch, i.e. the beginning of the tree.
            from_branch_id (str): The id of the branch that this branch is stemming from.
            from_tool_ids (list[str]): The ids of the tools that precede this branch being added (after the `from_branch_id` branch).
            status (str): The status message to be displayed when this branch is chosen.
        """
        if not root and description == "":
            raise ValueError("Description is required for non-root branches.")
        if not root and from_branch_id == "":
            raise ValueError(
                "`from_branch_id` is required for non-root branches. "
                "Set `root=True` to create a root branch or choose where this branch stems from."
            )
        if root and description != "":
            self.settings.logger.warning(f"Description is not used for root branches. ")
            description = ""

        if root and from_branch_id != "":
            self.settings.logger.warning(
                "`from_branch_id` is not used for root branches. "
                "(As this is the root branch, it does not stem from any other branch.)"
                "If you wish this to be stemming from a previous branch, set `root=False`."
            )
            from_branch_id = ""

        if status == "":
            status = f"Running {branch_id}..."

        decision_node = DecisionNode(
            id=branch_id,
            instruction=instruction,
            options={},
            root=root,
            logger=self.settings.logger,
            use_elysia_collections=self.use_elysia_collections,
        )
        self.decision_nodes[branch_id] = decision_node

        if not root:

            if from_tool_ids == []:
                self.decision_nodes[from_branch_id].add_option(
                    id=branch_id,
                    description=description,
                    inputs={},
                    action=None,
                    end=False,
                    status=status,
                    next=self.decision_nodes[branch_id],
                )

            else:

                current_decision_node = self.decision_nodes[from_branch_id]
                for from_tool_id in from_tool_ids:
                    if isinstance(current_decision_node, DecisionNode):
                        if from_tool_id not in current_decision_node.options:
                            raise ValueError(
                                f"Tool '{from_tool_id}' not found in branch '{from_branch_id}'. "
                                f"Available options are: {list(current_decision_node.options.keys())}"
                            )
                        current_decision_node = current_decision_node.options[
                            from_tool_id
                        ]["next"]

                new_branch_id = from_branch_id
                for from_tool_id in from_tool_ids:
                    new_branch_id += f".{from_tool_id}"

                # only create a new decision node if one doesn't exist here
                if new_branch_id not in self.decision_nodes:
                    decision_node = DecisionNode(
                        id=new_branch_id,
                        instruction=f"Choose one of the actions based on their descriptions and the user prompt.",
                        options={},
                        root=False,
                        logger=self.settings.logger,
                        use_elysia_collections=self.use_elysia_collections,
                    )
                    self.decision_nodes[new_branch_id] = decision_node

                    prev_branch_id = branch_id
                    for from_tool_id in from_tool_ids[:-1]:
                        prev_branch_id += f".{from_tool_id}"

                    self.decision_nodes[prev_branch_id].options[from_tool_ids[-1]][
                        "next"
                    ] = self.decision_nodes[new_branch_id]

                # add the tool to the new decision node
                self.decision_nodes[new_branch_id].add_option(
                    id=branch_id,
                    description=description,
                    inputs={},
                    action=None,
                    end=False,
                    status=status,
                    next=self.decision_nodes[branch_id],
                )

        if root and (self.root is not None):
            # replace root branch with this one
            self.decision_nodes[self.root] = decision_node
            self.settings.logger.debug(
                f"Replacing root branch '{self.root}' with '{branch_id}'."
            )
            old_root = self.root
            self.root = branch_id
            self.remove_branch(old_root)

        # reconstruct tree
        self._get_root()
        self.tree = {}
        self._construct_tree(self.root, self.tree)

    def remove_branch(self, branch_id: str) -> None:
        """
        Remove a branch from the tree.

        Args:
            branch_id (str): The id of the branch to remove
        """
        # Validate branch exists
        if branch_id not in self.decision_nodes:
            self.settings.logger.warning(
                f"Branch {branch_id} not found, nothing to remove."
            )
            return

        # Special handling for root node
        if (
            branch_id == self.root
            and sum(1 for node in self.decision_nodes.values() if node.root) == 1
        ):
            self.settings.logger.error(
                "Cannot remove root branch if there is only one root branch."
            )
            raise ValueError(
                "Cannot remove the root branch when there is only one root branch. "
                "Create a new root branch via .add_branch(..., root=True) first. "
                "(You could be trying to replace a root branch with the same ID as the one you are trying to remove. "
                "Try a different name for the new root branch.)"
            )

        for decision_node_id in self.decision_nodes:
            self.decision_nodes[decision_node_id].remove_option(branch_id)

        if branch_id in self.decision_nodes:
            del self.decision_nodes[branch_id]

        # reconstruct tree
        self._get_root()
        self.tree = {}
        self._construct_tree(self.root, self.tree)

    def view(
        self,
        indent: int = 0,
        prefix: str = "",
        max_width: int = 80,
        tree_dict: dict | None = None,
    ):
        """
        Format a tree dictionary into a nice hierarchical text representation.

        Args:
            tree_dict: The tree dictionary to format
            indent: Current indentation level
            prefix: Prefix for the current line (for tree structure visualization)
            max_width: Maximum width for text wrapping

        Returns:
            str: Formatted tree string
        """
        if tree_dict is None:
            tree_dict = self.tree

        result = []

        name = tree_dict.get("name", "Unknown")
        node_id = tree_dict.get("id", "")
        description = tree_dict.get("description", "")
        is_branch = tree_dict.get("branch", False)

        indent_str = "  " * indent
        node_line = (
            f"{indent_str}{prefix}📁 {name}"
            if is_branch
            else f"{indent_str}{prefix}🔧 {name}"
        )

        result.append(node_line)

        if description:
            desc_indent = len(indent_str) + 4  # Extra space for description
            available_width = max_width - desc_indent

            wrapped_desc = textwrap.fill(
                description,
                width=available_width,
                initial_indent="",
                subsequent_indent="",
            )

            for i, line in enumerate(wrapped_desc.split("\n")):
                if i == 0:
                    result.append(f"{indent_str}    💬 {line}")
                else:
                    result.append(f"{indent_str}       {line}")

            result.append("")

        options = tree_dict.get("options", {})
        if options:
            option_items = list(options.items())
            for i, (key, option) in enumerate(option_items):
                is_last = i == len(option_items) - 1
                child_prefix = "└── " if is_last else "├── "
                child_result = self.view(
                    indent + 1, child_prefix, max_width, tree_dict=option
                )
                result.append(child_result)

                if indent == 0 and not is_last:
                    result.append("")

        return "\n".join(result)

    @property
    def conversation_history(self):
        return self.tree_data.conversation_history

    @property
    def environment(self):
        return self.tree_data.environment

    async def create_conversation_title_async(self) -> str:
        """
        Create a title for the tree (async) using the base LM.
        Also assigns the `conversation_title` attribute to the tree.

        Returns:
            (str): The title for the tree.
        """
        with ElysiaKeyManager(self.settings):
            self.conversation_title = await create_conversation_title(
                self.tree_data.conversation_history, self.base_lm
            )
        return self.conversation_title

    def create_conversation_title(self) -> str:
        """
        Create a title for the tree using the base LM.
        Also assigns the `conversation_title` attribute to the tree.

        Returns:
            (str): The title for the tree.
        """
        return asyncio_run(self.create_conversation_title_async())

    async def get_follow_up_suggestions_async(
        self, context: str | None = None, num_suggestions: int = 2
    ) -> list[str]:
        """
        Get follow-up suggestions for the current user prompt via a base model LLM call.

        E.g., if the user asks "What was the most recent Github Issue?",
            and the results show a message from 'Jane Doe',
            the follow-up suggestions might be "What other issues did Jane Doe work on?"

        Args:
            context (str | None): A description of the type of follow-up questions to suggest
            num_suggestions (int): The number of follow-up suggestions to return (length of the list output)

        Returns:
            (list[str]): A list of follow-up suggestions
        """
        with ElysiaKeyManager(self.settings):
            suggestions = await get_follow_up_suggestions(
                self.tree_data,
                self.suggestions,
                self.base_lm,
                context=context,
                num_suggestions=num_suggestions,
            )
        if suggestions != []:
            self.settings.logger.debug(f"Follow-up suggestions: {suggestions}")
        else:
            self.settings.logger.error("No follow-up suggestions found.")

        self.suggestions.extend(suggestions)
        return suggestions

    def get_follow_up_suggestions(
        self,
        context: str | None = None,
        num_suggestions: int = 2,
    ) -> list[str]:
        """
        Get follow-up suggestions for the current user prompt via a base model LLM call (sync wrapper for get_follow_up_suggestions_async).

        E.g., if the user asks "What was the most recent Github Issue?",
            and the results show a message from 'Jane Doe',
            the follow-up suggestions might be "What other issues did Jane Doe work on?"

        Args:
            context (str | None): A description of the type of follow-up questions to suggest
            num_suggestions (int): The number of follow-up suggestions to return (length of the list output)

        Returns:
            (list[str]): A list of follow-up suggestions
        """
        return asyncio_run(
            self.get_follow_up_suggestions_async(context, num_suggestions)
        )

    def _update_conversation_history(self, role: str, message: str) -> None:
        if message != "":
            # If the first message, create a new message
            if len(self.tree_data.conversation_history) == 0:
                self.tree_data.update_list(
                    "conversation_history", {"role": role, "content": message}
                )
            # If the last message is from the same role, append to the content
            elif self.tree_data.conversation_history[-1]["role"] == role:
                if self.tree_data.conversation_history[-1]["content"].endswith(" "):
                    self.tree_data.conversation_history[-1]["content"] += message
                else:
                    self.tree_data.conversation_history[-1]["content"] += " " + message
            # Otherwise, create a new message
            else:
                self.tree_data.update_list(
                    "conversation_history", {"role": role, "content": message}
                )

    def _update_actions_called(self, result: Result, decision: Decision) -> None:
        if self.user_prompt not in self.actions_called:
            self.actions_called[self.user_prompt] = []
            self.actions_called[self.user_prompt].append(
                {
                    "name": decision.function_name,
                    "inputs": decision.function_inputs,
                    "reasoning": decision.reasoning,
                    "output": None,
                }
            )
        if not self.low_memory:
            self.actions_called[self.user_prompt][-1]["output"] = result.objects
        else:
            self.actions_called[self.user_prompt][-1]["output"] = []

    def _add_refs(self, objects: list[dict], tool_name: str, name: str) -> None:

        if (
            tool_name not in self.tree_data.environment.environment
            or name not in self.tree_data.environment.environment[tool_name]
        ):
            len_objects = 0
        else:
            len_objects = len(self.tree_data.environment.environment[tool_name][name])

        for i, obj in enumerate(objects):
            if "_REF_ID" not in obj:
                _REF_ID = f"{tool_name}_{name}_{len_objects}_{i}"
                obj["_REF_ID"] = _REF_ID

    def _update_environment(self, result: Result, decision: Decision) -> None:
        """
        Given a yielded result from an action or otherwise, update the environment.
        I.e. the items within the LLM knowledge base/prompt for future decisions/actions
        All Result subclasses have their .to_json() method added to the environment.
        As well, all Result subclasses have their llm_parse() method added to the tasks_completed.
        """

        # add to environment (store of retrieved/called objects)
        self.tree_data.environment.add(decision.function_name, result)

        # make note of which objects were retrieved _this session_ (for returning)
        if self.store_retrieved_objects:
            self.retrieved_objects.append(result.to_json(mapping=False))

        # add to log of actions called
        self.action_information.append(
            {
                "action_name": decision.function_name,
                **{key: value for key, value in result.metadata.items()},
            }
        )

        # add to tasks completed (parsed info / train of thought for LLM)
        self.tree_data.update_tasks_completed(
            prompt=self.user_prompt,
            task=decision.function_name,
            num_trees_completed=self.tree_data.num_trees_completed,
            reasoning=decision.reasoning,
            inputs=decision.function_inputs,
            parsed_info=result.llm_parse(),
            action=True,
        )

        # add to log of actions called
        self._update_actions_called(result, decision)

    def _add_error(self, function_name: str, error: Error) -> None:
        if function_name not in self.tree_data.errors:
            self.tree_data.errors[function_name] = []

        self.tree_data.update_tasks_completed(
            prompt=self.user_prompt,
            task=function_name,
            num_trees_completed=self.tree_data.num_trees_completed,
            error=True,
        )

        if error.feedback != "An unknown issue occurred.":
            self.tree_data.errors[function_name].append(
                "Avoidable error: "
                f"{error.feedback} "
                "(this error is likely to be solved by incorporating the feedback in a future tool call)"
            )
        else:
            self.tree_data.errors[function_name].append(
                "Unknown error: "
                f"{error.error_message} "
                "(this error is likely outside of your capacity to be solved - "
                "judge the error message based on other information and if it seems fixable, call this tool again "
                "if it is repeated, you may need to try something else or inform the user of the issue)"
            )

    async def _evaluate_result(
        self,
        result: Result | TreeUpdate | Error | TrainingUpdate | Text | Update,
        decision: Decision,
    ) -> tuple[dict | None, bool]:
        error = False

        if isinstance(result, Result):
            self._add_refs(result.objects, decision.function_name, result.name)
            self._update_environment(result, decision)

        if isinstance(result, TrainingUpdate):
            self.training_updates.append(result)
            return None, error

        if isinstance(result, Error):
            self._add_error(decision.function_name, result)
            if self.settings.LOGGING_LEVEL_INT <= 20:
                print(
                    Panel.fit(
                        (
                            result.error_message
                            if result.feedback == "An unknown issue occurred."
                            else result.feedback
                        ),
                        title="Error",
                        border_style="red",
                        padding=(1, 1),
                    )
                )
            error = True

        if isinstance(result, Text):
            self._update_conversation_history("assistant", result.text)
            if self.settings.LOGGING_LEVEL_INT <= 20:
                print(
                    Panel.fit(
                        result.text,
                        title="Assistant response",
                        border_style="cyan",
                        padding=(1, 1),
                    )
                )

        return (
            await self.returner(
                result,
                self.prompt_to_query_id[self.user_prompt],
            ),
            error,
        )

    async def _get_available_tools(
        self, current_decision_node: DecisionNode, client_manager: ClientManager
    ) -> tuple[list[str], list[tuple[str, str]]]:
        available_tools = []
        unavailable_tools = []
        for tool in current_decision_node.options.keys():
            if current_decision_node.options[tool]["action"] is None:
                available_tools.append(tool)
            elif "is_tool_available" in dir(self.tools[tool]) and await self.tools[
                tool
            ].is_tool_available(
                tree_data=self.tree_data,
                base_lm=self.base_lm,
                complex_lm=self.complex_lm,
                client_manager=client_manager,
            ):
                available_tools.append(tool)
            else:
                is_tool_available_doc = (
                    self.tools[tool].is_tool_available.__doc__.strip()
                    if self.tools[tool].is_tool_available.__doc__
                    else ""
                )
                unavailable_tools.append((tool, is_tool_available_doc))
        return available_tools, unavailable_tools

    def _get_successive_actions(
        self, successive_actions: dict, current_options: dict
    ) -> dict:

        for branch in current_options:
            successive_actions[branch] = {}
            if current_options[branch]["options"] != {}:
                successive_actions[branch] = self._get_successive_actions(
                    successive_actions[branch], current_options[branch]["options"]
                )
        return successive_actions

    def log_token_usage(self) -> None:
        if not self.low_memory:
            avg_input_base = self.tracker.get_average_input_tokens("base_lm")
            avg_output_base = self.tracker.get_average_output_tokens("base_lm")
            total_input_base = self.tracker.get_total_input_tokens("base_lm")
            total_output_base = self.tracker.get_total_output_tokens("base_lm")
            avg_input_complex = self.tracker.get_average_input_tokens("complex_lm")
            avg_output_complex = self.tracker.get_average_output_tokens("complex_lm")
            total_input_complex = self.tracker.get_total_input_tokens("complex_lm")
            total_output_complex = self.tracker.get_total_output_tokens("complex_lm")
            total_cost_base = self.tracker.get_total_cost("base_lm")
            total_cost_complex = self.tracker.get_total_cost("complex_lm")
            avg_cost_base = self.tracker.get_average_cost("base_lm")
            avg_cost_complex = self.tracker.get_average_cost("complex_lm")
            num_calls_base = self.tracker.get_num_calls("base_lm")
            num_calls_complex = self.tracker.get_num_calls("complex_lm")

            if num_calls_base > 0:
                self.settings.logger.debug(
                    f"Base Model Usage: \n"
                    f"  - Calls: [magenta]{num_calls_base}[/magenta]\n"
                    f"  - Input Tokens: [magenta]{total_input_base}[/magenta] (Avg. [magenta]{int(avg_input_base)}[/magenta] per call)\n"
                    f"  - Output Tokens: [cyan]{total_output_base}[/cyan] (Avg. [cyan]{int(avg_output_base)}[/cyan] per call)\n"
                    f"  - Total Cost: [yellow]${total_cost_base:.4f}[/yellow] (Avg. [yellow]${avg_cost_base:.4f}[/yellow] per call)\n"
                )
            else:
                self.settings.logger.debug(
                    f"Base Model Usage: [magenta]0[/magenta] calls"
                )
            if num_calls_complex > 0:
                self.settings.logger.debug(
                    f"Complex Model Usage: \n"
                    f"  - Calls: [magenta]{num_calls_complex}[/magenta]\n"
                    f"  - Input Tokens: [magenta]{total_input_complex}[/magenta] (Avg. [magenta]{int(avg_input_complex)}[/magenta] per call)\n"
                    f"  - Output Tokens: [cyan]{total_output_complex}[/cyan] (Avg. [cyan]{int(avg_output_complex)}[/cyan] per call)\n"
                    f"  - Total Cost: [yellow]${total_cost_complex:.4f}[/yellow] (Avg. [yellow]${avg_cost_complex:.4f}[/yellow] per call)\n"
                )
            else:
                self.settings.logger.debug(
                    f"Complex Model Usage: [magenta]0[/magenta] calls"
                )

    async def async_run(
        self,
        user_prompt: str,
        collection_names: list[str] = [],
        client_manager: ClientManager | None = None,
        training_route: str = "",
        query_id: str | None = None,
        close_clients_after_completion: bool = True,
        _first_run: bool = True,
        **kwargs,
    ) -> AsyncGenerator[dict | None, None]:
        """
        Async version of .run() for running Elysia in an async environment.
        See .run() for full documentation.
        """

        if client_manager is None:
            client_manager = ClientManager(
                wcd_url=self.settings.WCD_URL,
                wcd_api_key=self.settings.WCD_API_KEY,
                logger=self.settings.logger,
                client_timeout=None,
                **self.settings.API_KEYS,
            )

        # If training route is provided, split it into a list
        if training_route != "":
            route_list = training_route.split("/")
        else:
            route_list = []

        # Some initial steps if this is the first run (no recursion yet)
        if _first_run:

            self.settings.logger.debug(f"Style: {self.tree_data.atlas.style}")
            self.settings.logger.debug(
                f"Agent description: {self.tree_data.atlas.agent_description}"
            )
            self.settings.logger.debug(f"End goal: {self.tree_data.atlas.end_goal}")

            if query_id is None:
                query_id = str(uuid.uuid4())

            self.returner.add_prompt(user_prompt, query_id)

            # Reset the tree (clear temporary data specific to the last user prompt)
            self.soft_reset()

            check_base_lm_settings(self.settings)
            check_complex_lm_settings(self.settings)

            # Initialise some objects
            self.set_start_time()
            self.query_id_to_prompt[query_id] = user_prompt
            self.prompt_to_query_id[user_prompt] = query_id
            self.tree_data.set_property("user_prompt", user_prompt)
            self._update_conversation_history("user", user_prompt)
            self.user_prompt = user_prompt

            # check and start clients if not already started
            if client_manager.is_client:
                await client_manager.start_clients()

                # Initialise the collections
                if self.use_elysia_collections:
                    if collection_names == []:
                        async with client_manager.connect_to_async_client() as client:
                            collection_names = await retrieve_all_collection_names(
                                client
                            )
                    await self.set_collection_names(
                        collection_names,
                        client_manager,
                    )

            # If there are any empty branches, remove them (no tools attached to them)
            self._remove_empty_branches()

            if self.settings.LOGGING_LEVEL_INT <= 20:
                print(
                    Panel.fit(
                        user_prompt,
                        title="User prompt",
                        border_style="yellow",
                        padding=(1, 1),
                    )
                )

        # Start the tree at the root node
        if self.root is not None:
            current_decision_node: DecisionNode = self.decision_nodes[self.root]
        else:
            raise ValueError("No root node found!")

        # Loop through the tree until the end is reached
        while True:

            available_tools, unavailable_tools = await self._get_available_tools(
                current_decision_node, client_manager
            )

            if len(available_tools) == 0:
                self.settings.logger.error("No tools available to use!")
                raise ValueError(
                    "No tools available to use! "
                    "Check the tool definitions and the `is_tool_available` methods."
                )

            init_options = deepcopy(self.tree["options"])
            successive_actions = self._get_successive_actions(
                successive_actions={},
                current_options=init_options,
            )

            # Evaluate any tools which have hardcoded rules that have been met
            nodes_with_rules_met, rule_tool_inputs = await self._check_rules(
                current_decision_node.id, client_manager
            )

            if len(nodes_with_rules_met) > 0:
                for rule in nodes_with_rules_met:
                    rule_decision = Decision(rule, {}, "", False, False)
                    with ElysiaKeyManager(self.settings):
                        async for result in self.tools[rule](
                            tree_data=self.tree_data,
                            inputs=rule_tool_inputs[rule],
                            base_lm=self.base_lm,
                            complex_lm=self.complex_lm,
                            client_manager=client_manager,
                        ):
                            action_result, _ = await self._evaluate_result(
                                result, rule_decision
                            )
                            if action_result is not None:
                                yield action_result

            # If training route is provided, decide from the training route
            if len(route_list) > 0:
                self.settings.logger.debug(f"Route that will be used: {route_list}")

                (
                    self.current_decision,
                    training_route,
                ) = current_decision_node.decide_from_route(route_list)

                force_text_response = (
                    self.current_decision.function_name == "text_response"
                )

            # Under normal circumstances decide from the decision node
            else:
                self.tracker.start_tracking("decision_node")
                self.tree_data.set_current_task("elysia_decision_node")
                with ElysiaKeyManager(self.settings):
                    self.current_decision, results = await current_decision_node(
                        tree_data=self.tree_data,
                        base_lm=self.base_lm,
                        complex_lm=self.complex_lm,
                        available_tools=available_tools,
                        unavailable_tools=unavailable_tools,
                        successive_actions=successive_actions,
                        client_manager=client_manager,
                    )

                for result in results:
                    action_result, _ = await self._evaluate_result(
                        result, self.current_decision
                    )
                    if action_result is not None:
                        yield action_result

                self.tracker.end_tracking(
                    "decision_node",
                    "Decision Node",
                    self.base_lm if not self.low_memory else None,
                    self.complex_lm if not self.low_memory else None,
                )

                # Force text response (later) if model chooses end actions
                # but no response will be generated from the node, set flag now
                force_text_response = (
                    not current_decision_node.options[
                        self.current_decision.function_name
                    ]["end"]
                    and self.current_decision.end_actions
                )

            # Set default values for the function inputs for current call
            self.current_decision.function_inputs = self._get_function_inputs(
                self.current_decision.function_name,
                self.current_decision.function_inputs,
            )

            # end criteria, task picked is "text_response" or model chooses to end conversation
            completed = (
                self.current_decision.function_name == "text_response"
                or self.current_decision.end_actions
                or self.current_decision.impossible
                or self.tree_data.num_trees_completed > self.tree_data.recursion_limit
            )

            # assign action function
            action_fn: Tool | None = current_decision_node.options[
                self.current_decision.function_name
            ][
                "action"
            ]  # type: ignore

            # update the decision history
            self.decision_history[-1].append(self.current_decision.function_name)

            # print the current node information
            if self.settings.LOGGING_LEVEL_INT <= 20:
                print(
                    Panel.fit(
                        f"[bold]Node:[/bold] [magenta]{current_decision_node.id}[/magenta]\n"
                        f"[bold]Decision:[/bold] [green]{self.current_decision.function_name}[/green]\n"
                        f"[bold]Reasoning:[/bold] {self.current_decision.reasoning}\n",
                        title="Current Decision",
                        border_style="magenta",
                        padding=(1, 1),
                    )
                )

            self.tree_data.update_tasks_completed(
                prompt=self.user_prompt,
                task=self.current_decision.function_name,
                num_trees_completed=self.tree_data.num_trees_completed,
                reasoning=self.current_decision.reasoning,
                action=action_fn is not None,
            )

            # evaluate the action if this is not a branch
            if action_fn is not None:
                self.tracker.start_tracking(self.current_decision.function_name)
                self.tree_data.set_current_task(self.current_decision.function_name)
                successful_action = True
                with ElysiaKeyManager(self.settings):
                    async for result in action_fn(
                        tree_data=self.tree_data,
                        inputs=self.current_decision.function_inputs,
                        base_lm=self.base_lm,
                        complex_lm=self.complex_lm,
                        client_manager=client_manager,
                        **kwargs,
                    ):
                        action_result, error = await self._evaluate_result(
                            result, self.current_decision
                        )

                        if action_result is not None:
                            yield action_result

                        successful_action = not error and successful_action

                if not successful_action:
                    completed = (
                        False
                        or self.tree_data.num_trees_completed
                        > self.tree_data.recursion_limit
                    )

                if successful_action:
                    self.tree_data.clear_error(self.current_decision.function_name)

                self.tracker.end_tracking(
                    self.current_decision.function_name,
                    self.current_decision.function_name,
                    self.base_lm if not self.low_memory else None,
                    self.complex_lm if not self.low_memory else None,
                )

            yield (
                await self._evaluate_result(
                    TreeUpdate(
                        from_node=current_decision_node.id,
                        to_node=self.current_decision.function_name,
                        reasoning=(
                            self.current_decision.reasoning
                            if self.settings.BASE_USE_REASONING
                            else ""
                        ),
                        reset_tree=current_decision_node.options[
                            self.current_decision.function_name
                        ]["next"]
                        is None
                        and (not completed),
                    ),
                    self.current_decision,
                )
            )[0]

            # check if the current node is the end of the tree
            if (
                current_decision_node.options[self.current_decision.function_name][
                    "next"
                ]
                is None
                or completed
            ):
                break
            else:
                current_decision_node = current_decision_node.options[
                    self.current_decision.function_name
                ][
                    "next"
                ]  # type: ignore

        self.tree_data.num_trees_completed += 1

        # end of all trees
        if completed:

            # firstly, if we reached the end of a tree at a node that shouldn't be the end, call text response tool here to respond
            if (
                not current_decision_node.options[self.current_decision.function_name][
                    "end"
                ]
                or force_text_response
            ):
                with ElysiaKeyManager(self.settings):
                    async for result in self.tools["forced_text_response"](
                        tree_data=self.tree_data,
                        inputs={},
                        base_lm=self.base_lm,
                        complex_lm=self.complex_lm,
                    ):
                        action_result, _ = await self._evaluate_result(
                            result, self.current_decision
                        )
                        if action_result is not None:
                            yield action_result

            self.save_history(
                query_id=self.prompt_to_query_id[user_prompt],
                time_taken_seconds=time.time() - self.start_time,
            )

            yield await self.returner(
                Completed(), query_id=self.prompt_to_query_id[user_prompt]
            )

            self.settings.logger.debug(
                f"[bold green]Model identified overall goal as completed![/bold green]"
            )
            self.settings.logger.debug(
                f"Total time taken for decision tree: {time.time() - self.start_time:.2f} seconds"
            )
            self.settings.logger.debug(
                f"Decision Node Avg. Time: {self.tracker.get_average_time('decision_node'):.2f} seconds"
            )
            self.log_token_usage()

            avg_times = []
            for i, iteration in enumerate(self.decision_history):
                if iteration != []:
                    avg_times = [
                        (
                            f"  - {task} ([magenta]Avg. {self.tracker.get_average_time(task):.2f} seconds[/magenta])\n"
                            if task in self.tracker.trackers
                            else ""
                        )
                        for task in iteration
                    ]
                    self.settings.logger.debug(
                        f"Tasks completed (iteration {i+1}):\n" + "".join(avg_times)
                    )

            if close_clients_after_completion and client_manager.is_client:
                await client_manager.close_clients()

        # otherwise, end of the tree for this iteration, and recursively call process() to restart the tree
        else:
            self.settings.logger.debug(
                f"Model did [bold red]not[/bold red] yet complete overall goal! "
            )
            self.settings.logger.debug(
                f"Restarting tree (Recursion: {self.tree_data.num_trees_completed+1}/{self.tree_data.recursion_limit})..."
            )

            # recursive call to restart the tree since the goal was not completed
            self.decision_history.append([])
            async for result in self.async_run(
                user_prompt,
                collection_names,
                client_manager,
                training_route=training_route,
                query_id=query_id,
                _first_run=False,
            ):
                yield result

    def run(
        self,
        user_prompt: str,
        collection_names: list[str] = [],
        client_manager: ClientManager | None = None,
        training_route: str = "",
        query_id: str | None = None,
        close_clients_after_completion: bool = True,
    ) -> tuple[str, list[dict]]:
        """
        Run the Elysia decision tree.

        Args:
            user_prompt (str): The input from the user.
            collection_names (list[str]): The names of the collections to use.
                If not provided, Elysia will attempt to retrieve all collection names from the client.
            client_manager (ClientManager): The client manager to use.
                If not provided, a new ClientManager will be created.
            training_route (str): The route to use for training.
                Separate tools/branches you want to use with a "/".
                e.g. "query/text_response" will only use the "query" tool and the "text_response" tool, and end the tree there.
            query_id (str): The id of the query.
                Only necessary if you are hosting Elysia on a server with multiple users.
                If not provided, a new query id will be generated.
            close_clients_after_completion (bool): Whether to close the clients after the tree is completed.
                Leave as True for most use cases, but if you don't want to close the clients for the ClientManager, set to False.
                For example, if you are managing your own clients (e.g. in an app), you may want to set this to False.

        Returns:
            (str): The concatenation of all the responses from the tree.
            (list[dict]): The retrieved objects from the tree.
        """

        self.store_retrieved_objects = True

        async def run_process():
            async for result in self.async_run(
                user_prompt,
                collection_names,
                client_manager,
                training_route,
                query_id,
                close_clients_after_completion,
            ):
                pass
            return self.retrieved_objects

        async def run_with_live():
            console = Console()

            with console.status("[bold indigo]Thinking...") as status:
                async for result in self.async_run(
                    user_prompt,
                    collection_names,
                    client_manager,
                    training_route,
                    query_id,
                    close_clients_after_completion,
                ):
                    if (
                        result is not None
                        and "type" in result
                        and result["type"] == "status"
                        and isinstance(result["payload"], dict)
                        and "text" in result["payload"]
                    ):
                        payload: dict = result["payload"]  # type: ignore
                        status.update(f"[bold indigo]{payload['text']}")

            return self.retrieved_objects

        if self.settings.LOGGING_LEVEL_INT <= 20:
            yielded_results = asyncio_run(run_with_live())
        else:
            yielded_results = asyncio_run(run_process())

        text = self.tree_data.conversation_history[-1]["content"]

        return text, yielded_results

    def detailed_memory_usage(self) -> dict:
        """
        Returns a detailed breakdown of memory usage for all major objects in the Tree class.

        Returns:
            dict: Dictionary containing memory sizes (in bytes) for each major component
        """
        memory_usage = {}

        # Core data structures
        memory_usage["tree_data"] = asizeof.asizeof(self.tree_data)
        memory_usage["tree_data"] = asizeof.asizeof(self.tree_data)
        memory_usage["tree_data"] = asizeof.asizeof(self.tree_data)
        memory_usage["collection_data"] = asizeof.asizeof(
            self.tree_data.collection_data
        )

        # Decision nodes and tools
        memory_usage["decision_nodes"] = {
            node_id: asizeof.asizeof(node)
            for node_id, node in self.decision_nodes.items()
        }
        memory_usage["tools"] = {
            tool_name: asizeof.asizeof(tool) for tool_name, tool in self.tools.items()
        }

        # History and tracking
        memory_usage["decision_history"] = asizeof.asizeof(self.decision_history)
        memory_usage["history"] = asizeof.asizeof(self.history)
        memory_usage["training_updates"] = asizeof.asizeof(self.training_updates)
        memory_usage["action_information"] = asizeof.asizeof(self.action_information)

        # Mapping dictionaries
        memory_usage["query_mappings"] = {
            "query_id_to_prompt": asizeof.asizeof(self.query_id_to_prompt),
            "prompt_to_query_id": asizeof.asizeof(self.prompt_to_query_id),
        }

        # Calculate total
        memory_usage["total"] = sum(
            (v if isinstance(v, int) else sum(v.values()) if isinstance(v, dict) else 0)
            for v in memory_usage.values()
        )

        return memory_usage

    def export_to_json(self) -> dict:
        """
        Export the tree to a JSON object, to be used for loading the tree via import_from_json().

        Returns:
            (dict): The JSON object.
        """
        try:
            return {
                "user_id": self.user_id,
                "conversation_id": self.conversation_id,
                "conversation_title": self.conversation_title,
                "branch_initialisation": self.branch_initialisation,
                "use_elysia_collections": self.use_elysia_collections,
                "tree_index": self.tree_index,
                "store_retrieved_objects": self.store_retrieved_objects,
                "low_memory": self.low_memory,
                "tree_data": self.tree_data.to_json(remove_unserialisable=True),
                "settings": self.settings.to_json(),
                "tool_names": list(self.tools.keys()),
                "frontend_rebuild": self.returner.store,
            }
        except Exception as e:
            self.settings.logger.error(f"Error exporting tree to JSON: {str(e)}")
            raise e

    async def export_to_weaviate(
        self, collection_name: str, client_manager: ClientManager | None = None
    ) -> None:
        """
        Export the tree to a Weaviate collection.

        Args:
            collection_name (str): The name of the collection to export to.
            client_manager (ClientManager): The client manager to use.
                If not provided, a new ClientManager will be created from environment variables.
        """
        if client_manager is None:
            client_manager = ClientManager()
            close_after_use = True
        else:
            close_after_use = False

        async with client_manager.connect_to_async_client() as client:

            if not await client.collections.exists(collection_name):
                await client.collections.create(
                    collection_name,
                    vectorizer_config=wc.Configure.Vectorizer.none(),
                    inverted_index_config=wc.Configure.inverted_index(
                        index_timestamps=True
                    ),
                    properties=[
                        wc.Property(
                            name="user_id",
                            data_type=wc.DataType.TEXT,
                        ),
                        wc.Property(
                            name="conversation_id",
                            data_type=wc.DataType.TEXT,
                        ),
                        wc.Property(
                            name="tree",
                            data_type=wc.DataType.TEXT,
                        ),
                        wc.Property(
                            name="title",
                            data_type=wc.DataType.TEXT,
                        ),
                    ],
                )

            collection = client.collections.get(collection_name)

            json_data_str = json.dumps(self.export_to_json())

            uuid = generate_uuid5(self.conversation_id)

            if await collection.data.exists(uuid):
                await collection.data.update(
                    uuid=uuid,
                    properties={
                        "user_id": self.user_id,
                        "conversation_id": self.conversation_id,
                        "tree": json_data_str,
                        "title": self.conversation_title,
                    },
                )
                self.settings.logger.info(
                    f"Successfully updated existing tree in collection '{collection_name}' with id '{self.conversation_id}'"
                )
            else:
                await collection.data.insert(
                    uuid=uuid,
                    properties={
                        "user_id": self.user_id,
                        "conversation_id": self.conversation_id,
                        "tree": json_data_str,
                        "title": self.conversation_title,
                    },
                )
                self.settings.logger.info(
                    f"Successfully inserted new tree in collection '{collection_name}' with id '{self.conversation_id}'"
                )

        if close_after_use:
            await client_manager.close_clients()

    @classmethod
    def import_from_json(cls, json_data: dict) -> "Tree":
        """
        Import a tree from a JSON object, outputted by the export_to_json() method.

        Args:
            json_data (dict): The JSON object to import the tree from.

        Returns:
            (Tree): The new tree instance loaded from the JSON object.
        """
        settings = Settings.from_json(json_data["settings"])
        logger = settings.logger
        tree = cls(
            user_id=json_data["user_id"],
            conversation_id=json_data["conversation_id"],
            branch_initialisation=json_data["branch_initialisation"],
            style=json_data["tree_data"]["atlas"]["style"],
            agent_description=json_data["tree_data"]["atlas"]["agent_description"],
            end_goal=json_data["tree_data"]["atlas"]["end_goal"],
            low_memory=json_data["low_memory"],
            use_elysia_collections=json_data["use_elysia_collections"],
            settings=settings,
        )

        tree.returner.store = json_data["frontend_rebuild"]
        tree.tree_data = TreeData.from_json(json_data["tree_data"])
        tree.set_branch_initialisation(json_data["branch_initialisation"])

        # check tools
        for tool_name in json_data["tool_names"]:
            if tool_name not in tree.tools:
                logger.warning(
                    f"In saved tree, custom tool '{tool_name}' found. "
                    "This will not be loaded in the new tree. "
                    "You will need to add it to the tree manually."
                )

        return tree

    @classmethod
    async def import_from_weaviate(
        cls,
        collection_name: str,
        conversation_id: str,
        client_manager: ClientManager | None = None,
    ) -> "Tree":
        """
        Import a tree from a Weaviate collection.

        Args:
            collection_name (str): The name of the collection to import from.
            conversation_id (str): The id of the conversation to import.
            client_manager (ClientManager): The client manager to use.
                If not provided, a new ClientManager will be created from environment variables.

        Returns:
            (Tree): The tree object.
        """

        if client_manager is None:
            client_manager = ClientManager()
            close_after_use = True
        else:
            close_after_use = False

        async with client_manager.connect_to_async_client() as client:

            if not await client.collections.exists(collection_name):
                raise ValueError(
                    f"Collection '{collection_name}' does not exist in this Weaviate instance."
                )

            collection = client.collections.get(collection_name)
            uuid = generate_uuid5(conversation_id)
            # if not await collection.data.exists(uuid):
            #     raise ValueError(
            #         f"No tree found for conversation id '{conversation_id}' in collection '{collection_name}'."
            #     )

            response = await collection.query.fetch_object_by_id(uuid)

        if close_after_use:
            await client_manager.close_clients()

        if response is None:
            raise ValueError(
                f"No tree found for conversation id '{conversation_id}' in collection '{collection_name}'."
            )

        json_data_str = response.properties["tree"]
        json_data = json.loads(json_data_str)  # type: ignore

        return cls.import_from_json(json_data)

    def __call__(self, *args, **kwargs) -> tuple[str, list[dict]]:
        return self.run(*args, **kwargs)



import dspy
from elysia.tools.visualisation.objects import (
    BarChart,
    HistogramChart,
    ScatterOrLineChart,
)


class CreateBarChart(dspy.Signature):
    """
    Create one or more bar charts.

    Create a maximum of 9 bar charts.
    Each bar chart should have a maximum of 10 categories.
    Hence each chart should also have a maximum of 10 values per category.
    Pick the most relevant categories and values.
    """

    charts: list[BarChart] = dspy.OutputField(description="The bar chart to create.")
    overall_title: str = dspy.OutputField(
        description=(
            "If providing more than one chart, they will be displayed in a grid. "
            "This is the overall title for above the grid. "
            "Otherwise provide an empty string. "
        ),
    )


class CreateHistogramChart(dspy.Signature):
    """
    Create one or more histogram charts.

    Create a maximum of 9 histogram charts.
    Do not produce more than 50 values per histogram chart. Pick the most relevant values.
    """

    charts: list[HistogramChart] = dspy.OutputField(
        description="The histogram chart to create."
    )
    overall_title: str = dspy.OutputField(
        description=(
            "If providing more than one chart, they will be displayed in a grid. "
            "This is the overall title for above the grid. "
            "Otherwise provide an empty string. "
        ),
    )


class CreateScatterOrLineChart(dspy.Signature):
    """
    Create one or more scatter or line charts.

    Create a maximum of 9 scatter or line charts.
    Create a maximum of 50 points per scatter or line chart. Pick the most relevant points.

    A scatter or line chart can have multiple y-axis values, each with a different label.
    You can combine a line chart with a scatter chart by creating multiple
    """

    charts: list[ScatterOrLineChart] = dspy.OutputField(
        description="The scatter or line chart to create."
    )
    overall_title: str = dspy.OutputField(
        description=(
            "If providing more than one chart, they will be displayed in a grid. "
            "This is the overall title for above the grid. "
            "Otherwise provide an empty string. "
        ),
    )


# class CreateLineChart(dspy.Signature):
#     """
#     Create one or more line charts.

#     Create a maximum of 9 line charts.
#     Create a maximum of 50 points per line chart. Pick the most relevant points.
#     """

#     charts: list[ScatterOrLineChart] = dspy.OutputField(
#         description="The line chart to create."
#     )
#     overall_title: str = dspy.OutputField(
#         description=(
#             "If providing more than one chart, they will be displayed in a grid. "
#             "This is the overall title for above the grid. "
#             "Otherwise provide an empty string. "
#         ),
#     )



import os
import pytest
from elysia.config import Settings, configure

from elysia.config import settings as global_settings
from elysia.config import reset_settings

from elysia.tree import Tree


def reset_global_settings():
    reset_settings()


def test_setup():
    """
    Test that settings works
    """
    settings = Settings()

    # LLM should not be set by default
    assert settings.BASE_MODEL is None
    assert settings.COMPLEX_MODEL is None
    assert settings.BASE_PROVIDER is None
    assert settings.COMPLEX_PROVIDER is None
    assert settings.MODEL_API_BASE is None

    reset_global_settings()


def test_from_env_vars():
    """
    Test that the config can be loaded from env vars
    """
    os.environ["BASE_MODEL"] = os.getenv("BASE_MODEL", "gemini-2.5-flash")
    os.environ["COMPLEX_MODEL"] = os.getenv("COMPLEX_MODEL", "gemini-2.5-flash")
    os.environ["BASE_PROVIDER"] = os.getenv("BASE_PROVIDER", "openrouter/google")
    os.environ["COMPLEX_PROVIDER"] = os.getenv("COMPLEX_PROVIDER", "openrouter/google")
    os.environ["MODEL_API_BASE"] = os.getenv("MODEL_API_BASE", "https://api.test.com")
    os.environ["WCD_URL"] = os.getenv("WCD_URL", "http://test_url.com")
    os.environ["WCD_API_KEY"] = os.getenv("WCD_API_KEY", "test_key")
    os.environ["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY", "test_key_openai")
    os.environ["ANTHROPIC_API_KEY"] = os.getenv(
        "ANTHROPIC_API_KEY", "test_key_anthropic"
    )

    settings = Settings.from_env_vars()
    assert settings.BASE_MODEL == os.environ["BASE_MODEL"]
    assert settings.COMPLEX_MODEL == os.environ["COMPLEX_MODEL"]
    assert settings.BASE_PROVIDER == os.environ["BASE_PROVIDER"]
    assert settings.COMPLEX_PROVIDER == os.environ["COMPLEX_PROVIDER"]
    assert settings.MODEL_API_BASE == os.environ["MODEL_API_BASE"]
    assert settings.WCD_URL == os.environ["WCD_URL"]
    assert settings.WCD_API_KEY == os.environ["WCD_API_KEY"]
    assert settings.API_KEYS["openai_api_key"] == os.environ["OPENAI_API_KEY"]
    assert settings.API_KEYS["anthropic_api_key"] == os.environ["ANTHROPIC_API_KEY"]

    reset_global_settings()


def test_basic_configure_global():
    """
    Does using elysia.configure() change the global settings (for backend)
    """

    configure(base_model="gpt-4o-mini", base_provider="openai")
    assert global_settings.BASE_MODEL == "gpt-4o-mini"

    configure(base_model="gpt-4o", base_provider="openai")
    assert global_settings.BASE_MODEL != "gpt-4o-mini"
    assert global_settings.BASE_MODEL == "gpt-4o"

    configure(complex_model="claude-3-5-haiku-latest", complex_provider="anthropic")
    assert global_settings.COMPLEX_MODEL == "claude-3-5-haiku-latest"
    assert global_settings.COMPLEX_PROVIDER == "anthropic"

    reset_global_settings()


def test_basic_configure_local():
    """
    Does using elysia.configure() change the local settings (for backend)
    """
    settings = Settings()

    settings.configure(
        complex_model="claude-3-5-haiku-latest", complex_provider="anthropic"
    )
    assert settings.COMPLEX_MODEL == "claude-3-5-haiku-latest"
    assert settings.COMPLEX_PROVIDER == "anthropic"
    assert settings.COMPLEX_MODEL != global_settings.COMPLEX_MODEL
    assert settings.COMPLEX_PROVIDER != global_settings.COMPLEX_PROVIDER

    settings.configure(base_model="gpt-4o-mini", base_provider="openai")
    assert settings.BASE_MODEL == "gpt-4o-mini"
    assert settings.BASE_PROVIDER == "openai"
    assert settings.BASE_MODEL != global_settings.BASE_MODEL
    assert settings.BASE_PROVIDER != global_settings.BASE_PROVIDER

    settings.configure(base_model="gpt-4o", base_provider="openai")
    assert settings.BASE_MODEL != "gpt-4o-mini"
    assert settings.BASE_MODEL == "gpt-4o"
    assert settings.BASE_PROVIDER == "openai"
    assert settings.BASE_MODEL != global_settings.BASE_MODEL
    assert settings.BASE_PROVIDER != global_settings.BASE_PROVIDER

    reset_global_settings()


def test_model_keys():

    # use settings from smart setup
    settings = Settings()
    tree = Tree(settings=settings)

    # set the keys to wrong
    settings.configure(
        openrouter_api_key="wrong",
        base_model="gemini-2.0-flash-001",
        base_provider="openrouter/google",
        complex_model="gemini-2.0-flash-001",
        complex_provider="openrouter/google",
    )

    # explicitly set the adapter back to ChatAdapter
    from dspy.adapters import ChatAdapter
    import dspy

    with dspy.context(adapter=ChatAdapter()):
        # should error
        with pytest.raises(Exception):
            response, objects = tree("hi elly. use text response only")

        with pytest.raises(Exception):
            tree.create_conversation_title()

        with pytest.raises(Exception):
            tree.get_follow_up_suggestions()

        # missing keys
        settings = Settings()
        settings.configure(
            base_model="gemini-2.0-flash-001",
            base_provider="openrouter/google",
            complex_model="gemini-2.0-flash-001",
            complex_provider="openrouter/google",
        )

        tree = Tree(settings=settings)

        with pytest.raises(Exception):
            response, objects = tree("hi elly. use text response only")

        with pytest.raises(Exception):
            tree.create_conversation_title()

        with pytest.raises(Exception):
            tree.get_follow_up_suggestions()

    # now set the keys back to the .env
    settings.configure(
        openrouter_api_key=os.getenv("OPENROUTER_API_KEY"),
    )

    # should not error
    response, objects = tree("hi elly. use text response only")
    tree.create_conversation_title()
    tree.get_follow_up_suggestions()



from typing import Any
import dspy


class DecisionPrompt(dspy.Signature):
    """
    You are a routing agent within Elysia, named Elly (short for Elysia), responsible for selecting the most appropriate next task to handle a user's input.
    Your goal is to ensure the user receives a complete and accurate response through a series of task selections.
    You also respond to the user.

    Core Decision Process:
    1. Analyze the user's input prompt and available tasks
    2. Review completed tasks and their outcomes in previous_reasoning
    3. Check if current information satisfies the input prompt
    4. Select the most appropriate next task from available_tasks
    5. Determine if all possible actions have been exhausted

    Decision Rules:
    - Always select from available_tasks list only
    - Prefer tasks that directly progress toward answering the input prompt
    - Consider tree_count to avoid repetitive decisions

    Judge the task's possibility based on the user's prompt, the available actions, the previous errors and future possible actions (successive_actions).
    You are not designed with completing the prompt now, just choosing actions and later further actions from successive_actions which will complete the request.
    But you should judge the possibility of the request based on the available actions, previous errors and future possible actions.
    As you have access the descriptions of the available actions, you should use this to make your judgement.
    """

    # Regular input fields
    instruction: str = dspy.InputField(
        description="Specific guidance for this decision point that must be followed"
    )

    tree_count: str = dspy.InputField(
        description="""
        Current attempt number as "X/Y" where:
        - X = current attempt number
        - Y = maximum allowed attempts
        Consider ending the process as X approaches Y.
        """.strip()
    )

    # Task-specific input fields
    available_actions: list[dict] = dspy.InputField(
        description="""
        List of possible actions to choose from for this task only:
        {
            "[name]": {
                "function_name": [name of the function],
                "description": [task description],
                "future": [future information],
                "inputs": [inputs for the action]. A dictionary where the keys are the input names, and the values are the input descriptions.
            }
        }
        """.strip()
    )

    unavailable_actions: list[dict] = dspy.InputField(
        description="""
        List of actions that are unavailable to choose from for this task only:
        {
            "[name]": {
                "function_name": [name of the function],
                "available_at": [when or how the tool will be available],
            }
        }
        Do NOT pick tools from this list, it is there for information only.
        If you want to use this tool, you must complete the criteria in `available_at`.
        Sometimes the criteria in `available_at` could be something the user needs to complete.
        If you want to pick any tools from this list, you must first either complete the criteria in `available_at`,
        or inform the user what they need to do to complete the criteria.
        """.strip()
    )

    successive_actions: str = dspy.InputField(
        description="""
        Actions that stem from actions you can choose from.
        In the format:
        {
            "action_1": {
                "sub_action_1": {
                    "sub_sub_action_1": {},
                    "sub_sub_action_2": {},
                },
                "sub_action_2": {},
            },
            "action_2": {},
        }
        etc.
        Do NOT choose sub_actions for `function_name`, only choose actions from `available_actions`.
        Use this to inform your decision in how it will lead to future tools.
        """.strip()
    )

    previous_errors: list[dict] = dspy.InputField(
        description="""
        A list of errors from previous actions, organized by the function_name where each error occurred.
        Use this list to avoid repeating the same errors for those functions.
        If an error appears solvable, you should generally try the same tool again, as the error will be passed to the tool for handling.
        Err on the side of attempting to solve errors if there is a reasonable chance of success.
        Only avoid a tool if the error is clearly unsolvable or if a different tool is more appropriate.
        If no suitable tools remain and the task is impossible, set `impossible` to True and inform the user, possibly suggesting an alternative approach.
        """.strip(),
        format=list,
    )

    function_name: str = dspy.OutputField(
        description="""
        Select exactly one function name from available_actions that best advances toward answering the user's input prompt.
        You MUST select one function name exactly as written as it appears in the `function_name` field of the dictionary.
        You can select a function again in the future if you need to, you are allowed to choose the same function multiple times.
        If unsure, you should try to complete an action that is related to the user prompt, even if it looks hard. 
        However, do not continually choose the same action.
        Use the action descriptions (and an instruction, if given) to make your decision.
        Choose the action that will progress the user's request the most, even if it is not solved immediately.
        It may be a partial solution on route to a full solution.
        Use the `successive_actions` to inform your decision in how it will lead to future tools.
        Be aware that all the actions you can choose from now can be chosen again in future (as well as their sub-actions).

        IMPORTANT: only choose actions from `available_actions`, not `unavailable_actions`.
        """.strip()
    )

    function_inputs: dict[str, Any] = dspy.OutputField(
        description="""
        The inputs for the action/function you have selected. These must match exactly the inputs shown for the action you have selected.
        If it does not, the code will error.
        Choose based on the user prompt and the environment.
        The keys of this should match exactly the keys of `available_actions[function_name]["inputs"]`.
        Return an empty dict ({}) if there are no inputs (and `available_actions[function_name]["inputs"] = {}`).
        Follow the schema of the inputs for the action/function you have selected.
        """.strip(),
        format=dict,
    )

    end_actions: bool = dspy.OutputField(
        description="""
        Has the `end_goal` been achieved?
        Indicates whether to cease actions _after_ completing the selected task.
        Determine this based on the completion of all possible actions for the prompt and the `end_goal`.
        Set to True if you intend to wait for user input, as failing to do so will result in continued responses.
        Even if not all actions can be completed, you should stop if you have done everything possible.
        """.strip()
    )


class FollowUpSuggestionsPrompt(dspy.Signature):
    """
    Expert at suggesting engaging follow-up questions based on recent user interactions.
    Generate questions that showcase system capabilities and maintain user interest.
    Questions should be fun, creative, and designed to impress.

    Since you are suggesting _follow-up_ questions, you should not suggest questions that are too similar to the user's prompt.
    Instead, try to think of something that can connect different data sources together, or provide new insights that the user may not have thought of.
    These should be fully formed questions, that you think the system is capable of answering, that the user would be interested in, and that show off the system's capabilities.

    Use `data_information` to see if there are cross-overs between the already retrieved data (environment) and any other data sources.
    It is important you study these in depth so you do not create follow-up questions that would be impossible or very difficult to answer.

    Do not suggest questions that will require _complex_ calculations or reasoning, it should be mostly focused on following up on the already retrieved data,
    or following up by retrieving similar or new data that is related to the already retrieved data.

    You must ALWAYS return a list of strings (for `suggestions`), with each string being a follow-up question. Do not leave it empty.
    """

    user_prompt = dspy.InputField(description="The user's input prompt.")
    reference: dict[str, str] = dspy.InputField(
        desc="""
        Current state information like date/time to frame.
        """.strip(),
        format=str,
    )
    context: str = dspy.InputField(
        description="""
        Context for the suggestions.
        The context gives an overview of the system you are a part of, and the type of follow-up questions you should suggest.
        But, the questions should still be possible. If the context is impossible to create questions about, fall back to generic follow-up questions.
        """.strip(),
        format=str,
    )
    conversation_history: list[dict] = dspy.InputField(
        description="""
        Full conversation history between user and assistant, including generated information in environment field.
        Format: [{"role": "user"/"assistant", "content": message}]
        Use as context for future responses if non-empty.
        """.strip(),
        format=str,
    )
    environment: str = dspy.InputField(
        description="Information gathered from completed tasks. These are already shown to the user. Use for context, and to cross reference with data_information.",
        format=str,
    )
    data_information: dict[str, str | dict] = dspy.InputField(
        desc="""
        Collection metadata to inform query choices. Format per collection:
        {
            "name": data collection name,
            "length": object count,
            "summary": collection summary,
            "fields": [
                {
                    "name": field_name,
                    "groups": a dict with the value and count of each group.
                        a comprehensive list of all unique values that exist in the field.
                        if this is None, then no relevant groups were found,
                    "mean": average field length (tokens/list items),
                    "range": min/max length values,
                    "type": data type
                }
            ]
        }
        Use this to see if there are cross-overs between the currently retrieved data (environment) and any other collections, to help make an interesting question.
        Also use this to NOT suggest questions that would be impossible.
        This is IMPORTANT. Do not suggest questions that would be impossible, i.e., you can see the distinct categories between the collections, and look for overlap between the environment and these groups.
        If you can't find any overlap, do not assume there is any.
        Instead, suggest a question that would be a generic follow-up based on the environment/prompt.
        But if you do see overlap, suggest a question that would be a follow-up based on the overlap!
        """.strip(),
        format=str,
    )
    old_suggestions: list[str] = dspy.InputField(
        description="A list of questions that have already been suggested. Do not suggest the same question (or anything similar) as any items in this list.",
        format=str,
    )
    num_suggestions: int = dspy.InputField(
        description="The number of follow-up questions to suggest.",
        format=int,
    )
    suggestions: list[str] = dspy.OutputField(
        description=(
            "A list of follow up questions to the user's prompt. "
            "These should be punctual, short, around 10 words or so. "
            "This should be a list of questions of size `num_suggestions`."
            "Try to mix up the vocabulary and type of question between them."
        ),
        format=list,
    )


class TitleCreatorPrompt(dspy.Signature):
    """
    You are an expert at creating a title for a given conversation.
    The conversation could be a single user prompt, or a series of user prompts and assistant responses.
    Create a meaningful but punctual title for the conversation.
    """

    conversation: list[dict] = dspy.InputField(
        description="""
        The conversation to create a title for.
        This is a list of dictionaries, with each dictionary containing a `role` and `content` key.
        The `role` key can be either `user` or `assistant`.
        The `content` key contains the message content.
        """.strip(),
        format=list,
    )
    title: str = dspy.OutputField(
        description="""
        The title for the conversation. This is a single, short, succinct summary that describes the topic of the conversation.
        Base it primarily on the user's request, but also include should be a general summary of the entire conversation.
        It should not start with a verb, and should be a single sentence.
        It should be no more than 10 words.
        It should be a single sentence, not a question.
        Remember, this is a title, an extremely succinct summary of the whole conversation.
        """.strip()
    )



from typing import Any, Dict, List, Optional, Literal

from pydantic import BaseModel


class QueryData(BaseModel):
    user_id: str
    conversation_id: str
    query_id: str
    query: str
    collection_names: list[str]
    route: Optional[str] = ""
    mimick: Optional[bool] = False


class ViewPaginatedCollectionData(BaseModel):
    page_size: int
    page_number: int
    query: str = ""
    sort_on: Optional[str] = None
    ascending: bool = False
    filter_config: dict[str, Any] = {}


class InitialiseTreeData(BaseModel):
    low_memory: bool = False


class MetadataNamedVectorData(BaseModel):
    name: str
    enabled: Optional[bool] = None
    description: Optional[str] = None


class MetadataFieldData(BaseModel):
    name: str
    description: str


class UpdateCollectionMetadataData(BaseModel):
    named_vectors: Optional[List[MetadataNamedVectorData]] = None
    summary: Optional[str] = None
    mappings: Optional[dict[str, dict[str, str]]] = None
    fields: Optional[List[MetadataFieldData]] = None


class NERData(BaseModel):
    text: str


class TitleData(BaseModel):
    user_id: str
    conversation_id: str
    text: str


class ObjectRelevanceData(BaseModel):
    user_id: str
    conversation_id: str
    query_id: str
    objects: list[dict]


class ProcessCollectionData(BaseModel):
    user_id: str
    collection_name: str


class DebugData(BaseModel):
    conversation_id: str
    user_id: str


class AddFeedbackData(BaseModel):
    user_id: str
    conversation_id: str
    query_id: str
    feedback: int


class RemoveFeedbackData(BaseModel):
    user_id: str
    conversation_id: str
    query_id: str


class GetUserRequestsData(BaseModel):
    user_id: str


class InstantReplyData(BaseModel):
    user_id: str
    user_prompt: str


class FollowUpSuggestionsData(BaseModel):
    user_id: str
    conversation_id: str


# class BackendConfig(BaseModel):
#     settings: dict[str, Any]
#     style: str
#     agent_description: str
#     end_goal: str
#     branch_initialisation: str


class SaveConfigUserData(BaseModel):
    name: str
    config: dict[str, Any]
    frontend_config: dict[str, Any]
    default: bool


class SaveConfigTreeData(BaseModel):
    settings: Optional[dict[str, Any]] = None
    style: Optional[str] = None
    agent_description: Optional[str] = None
    end_goal: Optional[str] = None
    branch_initialisation: Optional[str] = None


class UpdateFrontendConfigData(BaseModel):
    config: dict[str, Any]


class AddToolToTreeData(BaseModel):
    tool_name: str
    branch_id: str
    from_tool_ids: list[str]


class RemoveToolFromTreeData(BaseModel):
    tool_name: str
    branch_id: str
    from_tool_ids: list[str]


class AddBranchToTreeData(BaseModel):
    id: str
    description: str
    instruction: str
    from_branch_id: str
    from_tool_ids: list[str]
    status: str
    root: bool


class RemoveBranchFromTreeData(BaseModel):
    id: str


class AvailableModelsData(BaseModel):
    user_id: str



import os
import pytest
import asyncio
import dspy
from dspy import LM
from elysia.objects import Result
from copy import deepcopy
from elysia import Tool
from elysia.tree.objects import TreeData
from elysia.config import Settings, configure
from elysia.tree.tree import Tree
from elysia.tools.text.text import TextResponse, CitedSummarizer
from elysia.tools.retrieval.query import Query
from elysia.tools.retrieval.aggregate import Aggregate
from elysia.objects import Response, Result

try:
    # get_ipython
    os.chdir("../..")
except NameError:
    pass

from elysia.util.client import ClientManager


class RetrieveProductData(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="retrieve_product_data",
            description="Retrieve product data from the product data collection.",
        )

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):
        yield Result(
            objects=[
                {
                    "product_id": "prod1",
                    "title": "Wireless Bluetooth Headphones",
                    "description": "High-quality wireless headphones with noise cancellation and 30-hour battery life.",
                    "tags": ["electronics", "audio", "wireless", "bluetooth"],
                    "price": 199.99,
                },
            ]
        )


class CheckResult(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="check_result",
            description="Check the result of the previous tool.",
        )

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):
        yield Response("Looks good to me!")


class SendEmail(Tool):
    def __init__(self, **kwargs):
        super().__init__(
            name="send_email",
            description="Send an email.",
            inputs={
                "email_address": {
                    "type": str,
                    "description": "The email address to send the email to.",
                    "required": True,
                }
            },
        )

    async def __call__(
        self, tree_data, inputs, base_lm, complex_lm, client_manager, **kwargs
    ):
        yield Response(f"Email sent to {inputs['email_address']}!")


class TestMultiBranch:

    @pytest.mark.asyncio
    async def test_incorrect_branch_id(self):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )

        # error if branch_id is not found
        with pytest.raises(ValueError):
            tree.add_tool(CitedSummarizer, branch_id="incorrect_branch_id")

        # no error if root is True
        tree.add_tool(CitedSummarizer, root=True)

    @pytest.mark.asyncio
    async def test_default_multi_branch(self):

        tree = Tree(
            low_memory=False,
            branch_initialisation="multi_branch",
            settings=Settings.from_smart_setup(),
        )

        assert len(tree.decision_nodes) > 1

        async for result in tree.async_run(
            user_prompt="Hello",
            collection_names=[],
        ):
            pass

    @pytest.mark.asyncio
    async def test_new_multi_branch(self):
        tree = Tree(
            low_memory=False,
            branch_initialisation="empty",
            settings=Settings.from_smart_setup(),
        )

        tree.add_branch(
            "respond_to_user",
            "choose between citing objects from the environment, or not citing any objects",
            "choose when responding to user",
            from_branch_id="base",
        )
        tree.add_branch(
            "search_for_objects",
            "choose between searching for objects from the environment, or aggregating objects from the environment",
            "choose when searching for objects",
            from_branch_id="base",
        )

        tree.add_tool(CitedSummarizer, branch_id="respond_to_user")
        tree.add_tool(TextResponse, branch_id="respond_to_user")
        tree.add_tool(Query, branch_id="search_for_objects")
        tree.add_tool(Aggregate, branch_id="search_for_objects")

        assert len(tree.decision_nodes) == 3

        async for result in tree.async_run(
            user_prompt="What was the most recent github issue?",
            collection_names=[],
        ):
            pass


def test_add_tool_with_stem_tool():
    tree = Tree(
        low_memory=False,
        branch_initialisation="empty",
        settings=Settings.from_smart_setup(),
    )

    tree.add_branch(
        branch_id="search",
        instruction="Search for information",
        description="Search for information",
        root=False,
        from_branch_id="base",
    )

    tree.add_tool(TextResponse, branch_id="base")
    tree.add_tool(RetrieveProductData, branch_id="search")
    tree.add_tool(
        CheckResult, branch_id="search", from_tool_ids=["retrieve_product_data"]
    )
    tree.add_tool(
        SendEmail,
        branch_id="search",
        from_tool_ids=["retrieve_product_data", "check_result"],
    )

    response, objects = tree(
        "Retrieve the product data for the product with the id 'prod1', then send an email to danny@weaviate.io with the product data",
        collection_names=[],
    )

    assert "Email sent to danny@weaviate.io!" in response
    assert "Looks good to me!" in response



from elysia.config import nlp


def named_entity_recognition(text: str):
    """
    Performs Named Entity Recognition using spaCy.
    Returns a list of entities with their labels, start and end positions.
    """
    try:
        doc = nlp(text)
        out = {"text": text, "entity_spans": [], "noun_spans": [], "error": ""}

        for ent in doc.ents:
            out["entity_spans"].append((ent.start_char, ent.end_char))

        # Get noun spans
        for token in doc:
            if token.pos_ == "NOUN":
                span = doc[token.i : token.i + 1]
                out["noun_spans"].append((span.start_char, span.end_char))

        return out

    except Exception as e:
        return {
            "text": text,
            "entity_spans": [],
            "noun_spans": [],
            "error": str(e),
        }



import asyncio
import uuid

from fastapi import APIRouter, Depends, WebSocket
from starlette.websockets import WebSocketDisconnect

from elysia.tree.tree import Tree
from elysia.api.core.log import logger
from elysia.api.dependencies.common import get_user_manager
from elysia.api.services.user import UserManager
from elysia.api.utils.websocket import help_websocket
from elysia.api.utils.ner import named_entity_recognition
from elysia.util.collection import retrieve_all_collection_names
from elysia.api.utils.default_payloads import error_payload

router = APIRouter()


def format_ner_response(text: str, user_id: str, conversation_id: str, query_id: str):
    response = named_entity_recognition(text)
    return {
        "type": "ner",
        "id": str(uuid.uuid4()),
        "user_id": user_id,
        "conversation_id": conversation_id,
        "query_id": query_id,
        "payload": response,
    }


async def format_title_response(
    tree: Tree, user_id: str, conversation_id: str, query_id: str
):

    try:
        title = await tree.create_conversation_title_async()
        return {
            "type": "title",
            "id": str(uuid.uuid4()),
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "payload": {
                "title": title,
                "error": "",
            },
        }
    except Exception as e:
        logger.exception(f"Error in format_title_response")
        return {
            "type": "title",
            "id": str(uuid.uuid4()),
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
            "payload": {
                "title": "",
                "error": str(e),
            },
        }


async def process(data: dict, websocket: WebSocket, user_manager: UserManager):
    logger.debug(f"/query API request received")
    logger.debug(f"User ID: {data['user_id']}")
    logger.debug(f"Conversation ID: {data['conversation_id']}")
    logger.debug(f"Query: {data['query']}")
    logger.debug(f"Query ID: {data['query_id']}")
    logger.debug(f"Training route: {data['route']}")
    logger.debug(f"Training mimick model: {data['mimick']}")
    logger.debug(f"Collection names: {data['collection_names']}")

    user = await user_manager.get_user_local(user_id=data["user_id"])

    try:
        # optional arguments
        if "route" in data:
            route = data["route"]
        else:
            route = ""

        # send ner response in advance
        await websocket.send_json(
            format_ner_response(
                text=data["query"],
                user_id=data["user_id"],
                conversation_id=data["conversation_id"],
                query_id=data["query_id"],
            )
        )

        async for yielded_result in user_manager.process_tree(
            user_id=data["user_id"],
            conversation_id=data["conversation_id"],
            query=data["query"],
            query_id=data["query_id"],
            training_route=route,
            collection_names=data["collection_names"],
        ):
            if asyncio.iscoroutine(yielded_result):
                yielded_result = await yielded_result
            try:
                if (
                    yielded_result is not None
                    and "type" in yielded_result
                    and yielded_result["type"] != "training_update"
                    and yielded_result["type"] != "timer"
                    and yielded_result["type"] != "completed"
                ):
                    await websocket.send_json(yielded_result)

                # before the completed, send title of conversation
                elif (
                    yielded_result is not None
                    and "type" in yielded_result
                    and yielded_result["type"] == "completed"
                ):
                    tree: Tree = await user_manager.get_tree(
                        user_id=data["user_id"],
                        conversation_id=data["conversation_id"],
                    )

                    # only send if it's the first prompt
                    if tree.tree_index == 0:
                        await websocket.send_json(
                            await format_title_response(
                                tree=tree,
                                user_id=data["user_id"],
                                conversation_id=data["conversation_id"],
                                query_id=data["query_id"],
                            )
                        )

                    # send the completed payload
                    await websocket.send_json(yielded_result)

            except WebSocketDisconnect:
                logger.info("Client disconnected during processing")
                break
            # Add a small delay between messages to prevent overwhelming
            await asyncio.sleep(0.005)
            # logger.debug(f"Sent message to client: {yielded_result}")

    except Exception as e:
        logger.exception(f"Error in /query API")

        if "conversation_id" in data:
            error = error_payload(
                text=f"{str(e)}",
                conversation_id=data["conversation_id"],
                query_id=data["query_id"],
            )
            await websocket.send_json(error)
        else:
            error = error_payload(
                text=f"{str(e)}",
                conversation_id="",
                query_id="",
            )
            await websocket.send_json(error)


# Process endpoint
@router.websocket("/query")
async def query_websocket(
    websocket: WebSocket, user_manager: UserManager = Depends(get_user_manager)
):
    """
    WebSocket endpoint for processing pipelines.
    Handles real-time communication for pipeline execution and status updates.
    """

    await help_websocket(websocket, lambda data, ws: process(data, ws, user_manager))



import os
from dotenv import set_key, load_dotenv
from rich import print

load_dotenv()

if "FIRST_START_ELYSIA" not in os.environ:
    print(
        "\n\n[bold green]Starting Elysia for the first time. This may take a minute to complete...[/bold green]\n\n"
    )
    set_key(".env", "FIRST_START_ELYSIA", "1")

import click
import uvicorn


@click.group()
def cli():
    """Main command group for Elysia."""
    pass


@cli.command()
@click.option(
    "--port",
    default=8000,
    help="FastAPI Port",
)
@click.option(
    "--host",
    default="localhost",
    help="FastAPI Host",
)
@click.option(
    "--reload",
    default=True,
    help="FastAPI Reload",
)
def start(port, host, reload):
    """
    Run the FastAPI application.
    """

    uvicorn.run(
        "elysia.api.app:app",
        host=host,
        port=port,
        reload=reload,
    )


if __name__ == "__main__":
    cli()



import pytest

from datetime import datetime

from elysia.api.routes.query import process
from elysia.api.dependencies.common import get_user_manager
from elysia.util.client import ClientManager
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.routes.feedback import (
    run_add_feedback,
    run_feedback_metadata,
    run_remove_feedback,
)

from elysia.api.api_types import (
    InitialiseTreeData,
    QueryData,
    AddFeedbackData,
    RemoveFeedbackData,
)


from fastapi.responses import JSONResponse
import json
from weaviate.util import generate_uuid5


def read_response(response: JSONResponse):
    return json.loads(response.body)


from elysia.api.core.log import logger, set_log_level

set_log_level("CRITICAL")


def read_response(response: JSONResponse):
    return json.loads(response.body)


class fake_websocket:
    results = []

    async def send_json(self, data: dict):
        self.results.append(data)


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )


@pytest.mark.asyncio
async def test_full_feedback_cycle():
    """
    Test the full feedback cycle.
    Add feedback, check metadata to see if it was added, remove feedback, check metadata to see if it was removed.
    """

    user_id = "test_user_full_feedback_cycle"
    conversation_id = "test_conversation_full_feedback_cycle"
    query_id = "test_query_full_feedback_cycle"
    user_manager = get_user_manager()
    websocket = fake_websocket()

    await initialise_user_and_tree(user_id, conversation_id)

    out = await process(
        QueryData(
            user_id=user_id,
            conversation_id=conversation_id,
            query="hi!",
            query_id=query_id,
            collection_names=[],
        ).model_dump(),
        websocket,
        user_manager,
    )

    response = await run_feedback_metadata(
        user_id=user_id,
        user_manager=user_manager,
    )

    assert read_response(response)["error"] == ""

    # record number of feedbacks
    num_feedbacks = read_response(response)["total_feedback"]

    # number of positives
    num_positives = read_response(response)["feedback_by_value"]["positive"]

    # number of dates
    date = datetime.now().strftime("%Y-%m-%d")
    if date in read_response(response)["feedback_by_date"]:
        num_feedbacks_on_this_date = read_response(response)["feedback_by_date"][date][
            "count"
        ]
    else:
        num_feedbacks_on_this_date = 0

    # check if the feedback exists already
    async with user_manager.users[user_id][
        "client_manager"
    ].connect_to_async_client() as client:
        feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")
        session_uuid = generate_uuid5(
            {
                "user_id": user_id,
                "conversation_id": conversation_id,
                "query_id": query_id,
            }
        )
        feedback_exists = await feedback_collection.data.exists(uuid=session_uuid)

    response = await run_add_feedback(
        AddFeedbackData(
            user_id=user_id,
            conversation_id=conversation_id,
            query_id=query_id,
            feedback=1,
        ),
        user_manager,
    )

    assert read_response(response)["error"] == ""

    response = await run_feedback_metadata(
        user_id=user_id,
        user_manager=user_manager,
    )

    assert read_response(response)["error"] == ""

    # number of positives should have increased by 1
    if not feedback_exists:
        assert (
            read_response(response)["feedback_by_value"]["positive"]
            == num_positives + 1
        )

        # number of feedbacks on this date should have increased by 1
        assert (
            read_response(response)["feedback_by_date"][date]["count"]
            == num_feedbacks_on_this_date + 1
        )

        # number of feedbacks should have increased by 1
        assert read_response(response)["total_feedback"] == num_feedbacks + 1
    else:
        assert read_response(response)["feedback_by_value"]["positive"] == num_positives
        assert (
            read_response(response)["feedback_by_date"][date]["count"]
            == num_feedbacks_on_this_date
        )
        assert read_response(response)["total_feedback"] == num_feedbacks

    response = await run_remove_feedback(
        RemoveFeedbackData(
            user_id=user_id,
            conversation_id=conversation_id,
            query_id=query_id,
        ),
        user_manager,
    )

    assert read_response(response)["error"] == ""

    response = await run_feedback_metadata(
        user_id=user_id,
        user_manager=user_manager,
    )

    assert read_response(response)["error"] == ""

    if not feedback_exists:
        # number of positives should have decreased by 1
        assert read_response(response)["feedback_by_value"]["positive"] == num_positives

        # number of feedbacks on this date should have decreased by 1 or been removed
        if date in read_response(response)["feedback_by_date"]:
            assert (
                read_response(response)["feedback_by_date"][date]["count"]
                == num_feedbacks_on_this_date
            )

        # number of feedbacks should have decreased by 1
        assert read_response(response)["total_feedback"] == num_feedbacks
    else:
        assert (
            read_response(response)["feedback_by_value"]["positive"]
            == num_positives - 1
        )
        assert read_response(response)["total_feedback"] == num_feedbacks - 1

        # number of feedbacks on this date should have decreased by 1 or been removed
        if date in read_response(response)["feedback_by_date"]:
            assert (
                read_response(response)["feedback_by_date"][date]["count"]
                == num_feedbacks_on_this_date - 1
            )



import uuid

# dspy requires a 'base' LM but this should not be used
import dspy
from pympler import asizeof
from logging import Logger

from typing import Callable, Union

from weaviate.util import generate_uuid5
from weaviate.classes.query import MetadataQuery, Sort, Filter

from elysia.objects import (
    Response,
    Result,
    Error,
    Text,
    Tool,
    Update,
    Status,
)

from elysia.tree.objects import TreeData
from elysia.util.objects import TrainingUpdate, TreeUpdate, FewShotExamples
from elysia.util.elysia_chain_of_thought import ElysiaChainOfThought
from elysia.util.parsing import format_datetime
from elysia.util.client import ClientManager
from elysia.tree.prompt_templates import (
    FollowUpSuggestionsPrompt,
    TitleCreatorPrompt,
    DecisionPrompt,
)
from elysia.tools.text.prompt_templates import TextResponsePrompt
from elysia.util.retrieve_feedback import retrieve_feedback


class ForcedTextResponse(Tool):
    """
    A tool that creates a new text response via a new LLM call.
    This is used in the decision tree when the tree reaches the end of its process, but no text is being displayed to the user.
    Then this tool is automaticaly called to create a new text response.
    """

    def __init__(self, **kwargs):
        super().__init__(
            name="final_text_response",
            description="",
            status="Writing response...",
            inputs={},
            end=True,
        )

    async def __call__(
        self,
        tree_data: TreeData,
        inputs: dict,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        client_manager: ClientManager | None = None,
        **kwargs,
    ):
        text_response = ElysiaChainOfThought(
            TextResponsePrompt,
            tree_data=tree_data,
            environment=True,
            tasks_completed=True,
            message_update=False,
        )

        output = await text_response.aforward(
            lm=base_lm,
        )

        yield Response(text=output.response)


class CopiedModule(dspy.Module):
    """
    A module that copies another module and adds a previous_feedbacks field to the signature.
    This is used to store the previous errored decision attempts for the decision node.
    This is one part of the feedback loop for the decision node, the other part is in the AssertedModule class.
    """

    def __init__(
        self,
        module: ElysiaChainOfThought,
        **kwargs,
    ):
        self.module = module

        feedback_desc = (
            "Pairs of INCORRECT previous attempts at this action, and the feedback received for each attempt. "
            "Judge what was incorrect in the previous attempts. "
            "Follow the feedback to improve your next attempt."
        )
        feedback_prefix = "${previous_feedback}"
        feedback_field: str = dspy.InputField(
            prefix=feedback_prefix, desc=feedback_desc
        )

        signature = self.module.predict.signature
        signature = signature.prepend(
            name="previous_feedbacks",
            field=feedback_field,
            type_=str,
        )

        self.module.predict.signature = signature  # type: ignore

    def _format_feedbacks(
        self, previous_feedbacks: list[str], previous_attempts: list[dict]
    ):
        feedbacks: list[str] = [
            f"ATTEMPT {i+1}:\n"
            + f"PREVIOUS INPUT: {str(attempt)}\n"
            + f"PREVIOUS FEEDBACK FOR THIS INPUT: {feedback}\n"
            for i, (attempt, feedback) in enumerate(
                zip(previous_attempts, previous_feedbacks)
            )
        ]
        return "\n".join(feedbacks)

    async def aforward(
        self,
        previous_feedbacks: list[str],
        previous_attempts: list[dict],
        **kwargs,
    ):
        pred = await self.module.acall(
            previous_feedbacks=self._format_feedbacks(
                previous_feedbacks, previous_attempts
            ),
            **kwargs,
        )
        return pred

    async def aforward_with_feedback_examples(
        self,
        previous_feedbacks: list[str],
        previous_attempts: list[dict],
        **kwargs,
    ):
        pred, uuids = await self.module.aforward_with_feedback_examples(
            previous_feedbacks=self._format_feedbacks(
                previous_feedbacks, previous_attempts
            ),
            **kwargs,
        )
        return pred, uuids


class AssertedModule(dspy.Module):
    """
    A module that calls another module until it passes an assertion function.
    This function returns a tuple of (asserted, feedback).
    If the assertion is false, the module is called again with the previous feedbacks and attempts.
    This is used to improve the module's performance and iteratively trying to get it to pass the assertion.
    """

    def __init__(
        self,
        module: ElysiaChainOfThought,
        assertion: Callable[[dict, dspy.Prediction], tuple[bool, str]],
        max_tries: int = 3,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.assertion = assertion
        self.module = module
        self.max_tries = max_tries

        self.previous_feedbacks = []
        self.previous_attempts = []

    def modify_signature_on_feedback(self, pred, feedback, **kwargs):
        self.previous_feedbacks.append(feedback)
        self.previous_attempts.append(pred.toDict())
        return CopiedModule(self.module.deepcopy(), **kwargs)

    async def aforward(self, **kwargs):

        pred = await self.module.acall(**kwargs)
        num_tries = 0

        asserted, feedback = self.assertion(kwargs, pred)

        while not asserted and num_tries <= self.max_tries:
            asserted_module = self.modify_signature_on_feedback(pred, feedback)
            pred = await asserted_module.aforward(
                previous_feedbacks=self.previous_feedbacks,
                previous_attempts=self.previous_attempts,
                **kwargs,
            )
            asserted, feedback = self.assertion(kwargs, pred)
            num_tries += 1

        return pred

    async def aforward_with_feedback_examples(self, **kwargs):
        pred, uuids = await self.module.aforward_with_feedback_examples(**kwargs)
        num_tries = 0

        asserted, feedback = self.assertion(kwargs, pred)

        while not asserted and num_tries <= self.max_tries:
            asserted_module = self.modify_signature_on_feedback(pred, feedback)
            pred, uuids = await asserted_module.aforward_with_feedback_examples(
                previous_feedbacks=self.previous_feedbacks,
                previous_attempts=self.previous_attempts,
                **kwargs,
            )
            asserted, feedback = self.assertion(kwargs, pred)
            num_tries += 1

        return pred, uuids


class Decision:
    """
    Simple decision object to store the decision made by the decision node.
    """

    def __init__(
        self,
        function_name: str,
        function_inputs: dict,
        reasoning: str,
        impossible: bool,
        end_actions: bool,
        last_in_tree: bool = False,
    ):
        self.function_name = function_name
        self.function_inputs = function_inputs
        self.reasoning = reasoning
        self.impossible = impossible
        self.end_actions = end_actions
        self.last_in_tree = last_in_tree


class DecisionNode:
    """
    A decision node is a node in the tree that makes a decision based on the available options.
    This class is essentially the executor of the decision node.
    """

    def __init__(
        self,
        id: str,
        instruction: str,
        options: dict[
            str, dict[str, Union[str, dict, bool, Tool, "DecisionNode", None]]
        ],
        root: bool = False,
        logger: Logger | None = None,
        use_elysia_collections: bool = True,
    ):
        self.id = id
        self.instruction = instruction
        self.options = options
        self.root = root
        self.logger = logger
        self.use_elysia_collections = use_elysia_collections

    def _get_options(self):
        return self.options

    def add_option(
        self,
        id: str,
        description: str,
        inputs: dict,
        action: Tool | None = None,
        end: bool = True,
        status: str = "",
        next: "DecisionNode | None" = None,
    ):
        if status == "":
            status = f"Running {id}..."

        self.options[id] = {
            "description": description,  # type: str
            "inputs": inputs,  # type: dict
            "action": action,  # type: Tool | None
            "end": end,  # type: bool
            "status": status,  # type: str
            "next": next,  # type: DecisionNode | None
        }

    def remove_option(self, id: str):
        if id in self.options:
            del self.options[id]

    def _options_to_json(self, available_tools: list[str]):
        """
        Options that get shown to the LLM.
        Remove any that are empty branches.
        """
        out = {}
        for node in self.options:
            if node not in available_tools:  # empty branch
                continue

            out[node] = {
                "function_name": node,
                "description": self.options[node]["description"],
            }

            if self.options[node]["inputs"] != {}:
                out[node]["inputs"] = self.options[node]["inputs"]

                for input_dict in out[node]["inputs"].values():
                    if hasattr(input_dict["type"], "model_json_schema"):
                        type_overwrite = f"A JSON object of the following properties: {input_dict['type'].model_json_schema()['properties']}"
                        if "$defs" in input_dict["type"].model_json_schema():
                            type_overwrite += f"\nWhere the values are: {input_dict['type'].model_json_schema()['$defs']}"
                        input_dict["type"] = type_overwrite
            else:
                out[node]["inputs"] = "No inputs are needed for this function."

        return out

    def _unavailable_options_to_json(self, unavailable_tools: list[tuple[str, str]]):
        """
        Options unavailable at this time and why.
        """
        out = {}
        for tool, reason in unavailable_tools:
            if reason == "":
                reason = "No reason provided."
            out[tool] = {
                "function_name": tool,
                "description": self.options[tool]["description"],
                "available_at": reason,
            }
        return out

    def decide_from_route(self, route: list[str]):
        possible_nodes = self._get_options()

        next_route = route[0]
        if next_route not in possible_nodes:
            raise Exception(
                f"Next node in training route ({next_route}) not in possible nodes ({possible_nodes})"
            )

        route = route[1:]
        completed = len(route) == 0

        return (
            Decision(
                function_name=next_route,
                reasoning=f"Decided to run {next_route} from route {route}",
                impossible=False,
                function_inputs={},
                end_actions=completed,
                last_in_tree=completed,
            ),
            "/".join(route),
        )

    async def load_model_from_examples(
        self,
        decision_executor: dspy.Module,
        client_manager: ClientManager,
        user_prompt: str,
    ) -> dspy.Module:

        examples = await retrieve_feedback(client_manager, user_prompt, "decision")

        if len(examples) == 0:
            return decision_executor

        optimizer = dspy.LabeledFewShot(k=10)
        compiled_executor = optimizer.compile(decision_executor, trainset=examples)
        return compiled_executor

    def _tool_assertion(self, kwargs, pred):
        return (
            pred.function_name in self.options,
            f"You picked the action `{pred.function_name}` - that is not in `available_actions`! "
            f"Your output MUST be one of the following: {list(self.options.keys())}",
        )

    async def __call__(
        self,
        tree_data: TreeData,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        available_tools: list[str],
        unavailable_tools: list[tuple[str, str]],
        successive_actions: dict,
        client_manager: ClientManager,
        **kwargs,
    ) -> tuple[Decision, list[TrainingUpdate | Status | Response | FewShotExamples]]:

        available_options = self._options_to_json(available_tools)
        unavailable_options = self._unavailable_options_to_json(unavailable_tools)

        if len(available_options) == 0:
            raise RuntimeError(
                "No available tools to call! Make sure you have added some tools to the tree. "
                "Or the .is_tool_available() method is returning True for at least one tool."
            )

        if self.logger:
            self.logger.debug(f"Available options: {list(available_options.keys())}")

        one_choice = (
            all(
                option["inputs"] == "No inputs are needed for this function."
                for option in available_options.values()
            )
            and len(available_options) == 1
        )

        if not one_choice:
            decision_module = ElysiaChainOfThought(
                DecisionPrompt,
                tree_data=tree_data,
                environment=True,
                collection_schemas=self.use_elysia_collections,
                tasks_completed=True,
                message_update=True,
                reasoning=tree_data.settings.BASE_USE_REASONING,
            )

            decision_executor = AssertedModule(
                decision_module,
                assertion=self._tool_assertion,
                max_tries=2,
            )

            if tree_data.settings.USE_FEEDBACK:
                if not client_manager.is_client:
                    raise ValueError(
                        "A Weaviate conneciton is required for the experimental `use_feedback` method. "
                        "Please set the WCD_URL and WCD_API_KEY in the settings. "
                        "Or, set `use_feedback` to False."
                    )

                output, uuids = await decision_executor.aforward_with_feedback_examples(
                    feedback_model="decision",
                    client_manager=client_manager,
                    base_lm=base_lm,
                    complex_lm=complex_lm,
                    instruction=self.instruction,
                    tree_count=tree_data.tree_count_string(),
                    available_actions=available_options,
                    unavailable_actions=unavailable_options,
                    successive_actions=successive_actions,
                    num_base_lm_examples=3,
                    return_example_uuids=True,
                )
            else:
                output = await decision_executor.aforward(
                    instruction=self.instruction,
                    tree_count=tree_data.tree_count_string(),
                    available_actions=available_options,
                    unavailable_actions=unavailable_options,
                    successive_actions=successive_actions,
                    lm=base_lm,
                )

            if output.function_name.startswith("'") and output.function_name.endswith(
                "'"
            ):
                output.function_name = output.function_name[1:-1]
            elif output.function_name.startswith('"') and output.function_name.endswith(
                '"'
            ):
                output.function_name = output.function_name[1:-1]
            elif output.function_name.startswith("`") and output.function_name.endswith(
                "`"
            ):
                output.function_name = output.function_name[1:-1]

            if output.function_name not in available_tools:
                raise Exception(
                    f"Model picked an action `{output.function_name}` that is not in the available tools: {available_tools}"
                )

            decision = Decision(
                output.function_name,
                output.function_inputs,
                output.reasoning if tree_data.settings.BASE_USE_REASONING else "",
                output.impossible,
                output.end_actions and bool(self.options[output.function_name]["end"]),
            )

            results = [
                TrainingUpdate(
                    module_name="decision",
                    inputs=tree_data.to_json(),
                    outputs={k: v for k, v in output.__dict__["_store"].items()},
                ),
                Status(str(self.options[output.function_name]["status"])),
            ]

            if output.function_name != "text_response":
                results.append(Response(output.message_update))

            if tree_data.settings.USE_FEEDBACK and len(uuids) > 0:
                results.append(FewShotExamples(uuids))

        else:
            decision = Decision(
                function_name=list(available_options.keys())[0],
                reasoning=f"Only one option available: {list(available_options.keys())[0]} (and no function inputs are needed).",
                impossible=False,
                function_inputs={},
                end_actions=(
                    bool(self.options[list(available_options.keys())[0]]["end"])
                    and self.options[list(available_options.keys())[0]]["next"] is None
                ),
            )

            results: list[TrainingUpdate | Status | Response | FewShotExamples] = [
                Status(str(self.options[list(available_options.keys())[0]]["status"])),
            ]

        return decision, results

    def detailed_memory_usage(self) -> dict:
        """
        Returns a detailed breakdown of memory usage for all objects in the DecisionNode.

        Returns:
            dict: Dictionary containing memory sizes (in bytes) for each component
        """
        memory_usage = {}

        # Core attributes
        memory_usage["id"] = asizeof.asizeof(self.id)
        memory_usage["instruction"] = asizeof.asizeof(self.instruction)
        memory_usage["root"] = asizeof.asizeof(self.root)

        # Decision-related data - break down each option
        memory_usage["options"] = {}
        for option_id, option in self.options.items():
            memory_usage["options"][option_id] = {
                "description": asizeof.asizeof(option["description"]),
                "inputs": asizeof.asizeof(option["inputs"]),
                "action": asizeof.asizeof(option["action"]),
                "end": asizeof.asizeof(option["end"]),
                "status": asizeof.asizeof(option["status"]),
                "next": asizeof.asizeof(option["next"]),
            }

        # Calculate total
        memory_usage["total"] = sum(
            (
                v
                if isinstance(v, int)
                else (
                    sum(
                        (
                            sv
                            if isinstance(sv, int)
                            else sum(sv.values()) if isinstance(sv, dict) else 0
                        )
                        for sv in v.values()
                    )
                    if isinstance(v, dict)
                    else 0
                )
            )
            for v in memory_usage.values()
        )

        return memory_usage


class TreeReturner:
    """
    Class to parse the output of the tree to the frontend.
    """

    def __init__(
        self,
        user_id: str,
        conversation_id: str,
        tree_index: int = 0,
    ):
        self.user_id = user_id
        self.conversation_id = conversation_id
        self.tree_index = tree_index
        self.store = []

    def set_tree_index(self, tree_index: int):
        self.tree_index = tree_index

    def clear_store(self):
        self.store = []

    def add_prompt(self, prompt: str, query_id: str):
        self.store.append(
            {
                "type": "user_prompt",
                "id": str(uuid.uuid4()),
                "user_id": self.user_id,
                "conversation_id": self.conversation_id,
                "query_id": query_id,
                "payload": {
                    "prompt": prompt,
                },
            }
        )

    async def __call__(
        self,
        result: Result | TreeUpdate | Update | Text | Error,
        query_id: str,
    ) -> dict[str, str | dict] | None:

        if isinstance(result, (Update, Text, Result, Error)):
            payload = await result.to_frontend(
                self.user_id, self.conversation_id, query_id
            )
            self.store.append(payload)
            return payload

        if isinstance(result, TreeUpdate):
            payload = await result.to_frontend(
                self.user_id,
                self.conversation_id,
                query_id,
                self.tree_index,
            )
            self.store.append(payload)
            return payload


async def create_conversation_title(conversation: list[dict], lm: dspy.LM):
    title_creator = dspy.Predict(TitleCreatorPrompt)
    title = await title_creator.aforward(
        conversation=conversation,
        lm=lm,
    )
    return title.title


async def get_follow_up_suggestions(
    tree_data: TreeData,
    current_suggestions: list[str],
    lm: dspy.LM,
    num_suggestions: int = 2,
    context: str | None = None,
):

    # load dspy model for suggestor
    follow_up_suggestor = dspy.Predict(FollowUpSuggestionsPrompt)

    if context is None:
        context = (
            "System context: You are an agentic RAG service querying or aggregating information from Weaviate collections via filtering, "
            "sorting, multi-collection queries, summary statistics. "
            "Uses tree-based approach with specialized agents and decision nodes for dynamic information retrieval, "
            "summary generation, and real-time results display while maintaining natural conversation flow. "
            "Create questions that are natural follow-ups to the user's prompt, "
            "which they may find interesting or create relevant insights to the already retrieved data. "
            "Or, questions which span across other collections, but are still relevant to the user's prompt."
        )

    # get prediction
    prediction = await follow_up_suggestor.aforward(
        user_prompt=tree_data.user_prompt,
        reference=tree_data.atlas.datetime_reference,
        conversation_history=tree_data.conversation_history,
        environment=tree_data.environment.environment,
        data_information=tree_data.output_collection_metadata(with_mappings=False),
        old_suggestions=current_suggestions,
        context=context,
        num_suggestions=num_suggestions,
        lm=lm,
    )

    return prediction.suggestions


async def get_saved_trees_weaviate(
    collection_name: str,
    client_manager: ClientManager | None = None,
    user_id: str | None = None,
):
    """
    Get all saved trees from a Weaviate collection.

    Args:
        collection_name (str): The name of the collection to get the trees from.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created from environment variables.
        user_id (str): The user ID to get the trees from.
            If not provided, the trees will be retrieved from the collection without any filters.

    Returns:
        dict: A dictionary of tree UUIDs and their titles.
    """
    if client_manager is None:
        client_manager = ClientManager()
        close_after_use = True
    else:
        close_after_use = False

    if user_id is not None:
        user_id_filter = Filter.by_property("user_id").equal(user_id)
    else:
        user_id_filter = None

    async with client_manager.connect_to_async_client() as client:

        if not await client.collections.exists(collection_name):
            return {}

        collection = client.collections.get(collection_name)

        len_collection = (
            await collection.aggregate.over_all(
                total_count=True,
                filters=user_id_filter,
            )
        ).total_count

        response = await collection.query.fetch_objects(
            limit=len_collection,
            sort=Sort.by_update_time(ascending=False),
            return_metadata=MetadataQuery(last_update_time=True),
            filters=user_id_filter,
        )

    if close_after_use:
        await client_manager.close_clients()

    trees = {
        obj.properties["conversation_id"]: {
            "title": obj.properties["title"],
            "last_update_time": format_datetime(obj.metadata.last_update_time),
        }
        for obj in response.objects
    }

    return trees


async def delete_tree_from_weaviate(
    conversation_id: str,
    collection_name: str,
    client_manager: ClientManager | None = None,
):
    """
    Delete a tree from a Weaviate collection.

    Args:
        conversation_id (str): The conversation ID of the tree to delete.
        collection_name (str): The name of the collection to delete the tree from.
        client_manager (ClientManager): The client manager to use.
            If not provided, a new ClientManager will be created from environment variables.
    """

    if client_manager is None:
        client_manager = ClientManager()

    async with client_manager.connect_to_async_client() as client:
        collection = client.collections.get(collection_name)
        uuid = generate_uuid5(conversation_id)
        if await collection.data.exists(uuid):
            await collection.data.delete_by_id(uuid)



from dspy import LM, configure, ChatAdapter
from dspy.utils import DummyLM
from contextlib import contextmanager
from pydantic_core import PydanticUndefined


class DummyAdapter(ChatAdapter):

    def __init__(
        self, callbacks: list | None = None, use_native_function_calling: bool = True
    ):
        super().__init__(
            callbacks=callbacks, use_native_function_calling=use_native_function_calling
        )

    def __call__(
        self,
        lm: LM,
        lm_kwargs,
        signature,
        demos,
        inputs,
    ):

        categories = []
        defaults = {}

        # set some specific defaults for different inputs

        # for decision node
        if "available_actions" in inputs:
            defaults["function_name"] = list(inputs["available_actions"].keys())[0]

        # for preprocessing return type assignment
        if "possible_return_types" in inputs:
            defaults["return_types"] = [list(inputs["possible_return_types"].keys())[0]]

        # for field mapping
        if "input_data_fields" in inputs and "output_data_fields" in inputs:
            defaults["field_mapping"] = {
                in_field: inputs["output_data_fields"][0]
                for in_field in inputs["input_data_fields"]
            }

        # otherwise, try to find the default value from the signature to ensure type checking
        for field_name, field in signature.model_fields.items():
            if field.json_schema_extra.get("__dspy_field_type") == "output":

                categories.append(field_name)

                if field_name not in defaults:
                    if field.default is not PydanticUndefined:
                        defaults[field_name] = field.default
                    else:
                        try:
                            defaults[field_name] = field.annotation()
                        except Exception as e:
                            try:
                                defaults[field_name] = field.annotation()
                            except Exception as e:
                                try:
                                    defaults[field_name] = field.json_schema_extra[
                                        "format"
                                    ]()
                                except Exception as e:
                                    defaults[field_name] = "Test!"

        return super().__call__(
            DummyLM([{cat: defaults[cat] for cat in categories}]),
            lm_kwargs,
            signature,
            demos,
            inputs,
        )

    async def acall(
        self,
        lm: LM,
        lm_kwargs,
        signature,
        demos,
        inputs,
    ):

        categories = []
        defaults = {}

        # set some specific defaults for different inputs

        # for decision node
        if "available_actions" in inputs:
            defaults["function_name"] = list(inputs["available_actions"].keys())[0]

        # for preprocessing return type assignment
        if "possible_return_types" in inputs:
            defaults["return_types"] = [list(inputs["possible_return_types"].keys())[0]]

        # for field mapping
        if "input_data_fields" in inputs and "output_data_fields" in inputs:
            defaults["field_mapping"] = {
                in_field: inputs["output_data_fields"][0]
                for in_field in inputs["input_data_fields"]
            }

        # otherwise, try to find the default value from the signature to ensure type checking
        for field_name, field in signature.model_fields.items():
            if field.json_schema_extra.get("__dspy_field_type") == "output":

                categories.append(field_name)

                if field_name not in defaults:
                    if field.default is not PydanticUndefined:
                        defaults[field_name] = field.default
                    else:
                        try:
                            defaults[field_name] = field.annotation()
                        except Exception as e:
                            try:
                                defaults[field_name] = field.annotation()
                            except Exception as e:
                                try:
                                    defaults[field_name] = field.json_schema_extra[
                                        "format"
                                    ]()
                                except Exception as e:
                                    defaults[field_name] = "Test!"

        return await super().acall(
            DummyLM([{cat: defaults[cat] for cat in categories}]),
            lm_kwargs,
            signature,
            demos,
            inputs,
        )



import pytest
from elysia.api.dependencies.common import get_user_manager


@pytest.fixture(scope="session", autouse=True)
async def cleanup_clients(request):
    yield
    user_manager = get_user_manager()
    await user_manager.close_all_clients()
